# -*- coding: utf-8 -*-
"""add batch status

Revision ID: 043d4bafc0be
Revises: 4d2c5bdfecff
Create Date: 2017-06-21 12:44:38.709412

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '043d4bafc0be'
down_revision = u'4d2c5bdfecff'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


def upgrade():

    # batch.status_code, batch.status_text
    op.add_column('batch_handheld', sa.Column('status_code', sa.Integer(), nullable=True))
    op.add_column('batch_handheld', sa.Column('status_text', sa.String(length=255), nullable=True))
    op.add_column('batch_inventory', sa.Column('status_code', sa.Integer(), nullable=True))
    op.add_column('batch_inventory', sa.Column('status_text', sa.String(length=255), nullable=True))
    op.add_column('batch_pricing', sa.Column('status_code', sa.Integer(), nullable=True))
    op.add_column('batch_pricing', sa.Column('status_text', sa.String(length=255), nullable=True))
    op.add_column('label_batch', sa.Column('status_code', sa.Integer(), nullable=True))
    op.add_column('label_batch', sa.Column('status_text', sa.String(length=255), nullable=True))
    op.add_column('purchase_batch', sa.Column('status_code', sa.Integer(), nullable=True))
    op.add_column('purchase_batch', sa.Column('status_text', sa.String(length=255), nullable=True))
    op.add_column('vendor_catalog', sa.Column('status_code', sa.Integer(), nullable=True))
    op.add_column('vendor_catalog', sa.Column('status_text', sa.String(length=255), nullable=True))
    op.add_column('vendor_invoice', sa.Column('status_code', sa.Integer(), nullable=True))
    op.add_column('vendor_invoice', sa.Column('status_text', sa.String(length=255), nullable=True))


def downgrade():

    # batch.status_code, batch.status_text
    op.drop_column('vendor_invoice', 'status_text')
    op.drop_column('vendor_invoice', 'status_code')
    op.drop_column('vendor_catalog', 'status_text')
    op.drop_column('vendor_catalog', 'status_code')
    op.drop_column('purchase_batch', 'status_text')
    op.drop_column('purchase_batch', 'status_code')
    op.drop_column('label_batch', 'status_text')
    op.drop_column('label_batch', 'status_code')
    op.drop_column('batch_pricing', 'status_text')
    op.drop_column('batch_pricing', 'status_code')
    op.drop_column('batch_inventory', 'status_text')
    op.drop_column('batch_inventory', 'status_code')
    op.drop_column('batch_handheld', 'status_text')
    op.drop_column('batch_handheld', 'status_code')



# -*- coding: utf-8; -*-
"""rename batch_vendorinvoice

Revision ID: 117236288811
Revises: 89e2406859be
Create Date: 2021-02-02 13:36:28.621440

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '117236288811'
down_revision = '89e2406859be'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


def upgrade():

    # remove references to `vendor_invoice`
    op.drop_constraint('vendor_invoice_row_fk_batch', 'vendor_invoice_row', type_='foreignkey')

    # batch_vendorinvoice
    op.drop_constraint('vendor_invoice_fk_created_by', 'vendor_invoice', type_='foreignkey')
    op.drop_constraint('vendor_invoice_fk_cognized_by', 'vendor_invoice', type_='foreignkey')
    op.drop_constraint('vendor_invoice_fk_executed_by', 'vendor_invoice', type_='foreignkey')
    op.drop_constraint('vendor_invoice_fk_vendor', 'vendor_invoice', type_='foreignkey')
    op.rename_table('vendor_invoice', 'batch_vendorinvoice')
    op.create_foreign_key('batch_vendorinvoice_fk_created_by',
                          'batch_vendorinvoice', 'user',
                          ['created_by_uuid'], ['uuid'])
    op.create_foreign_key('batch_vendorinvoice_fk_cognized_by',
                          'batch_vendorinvoice', 'user',
                          ['cognized_by_uuid'], ['uuid'])
    op.create_foreign_key('batch_vendorinvoice_fk_executed_by',
                          'batch_vendorinvoice', 'user',
                          ['executed_by_uuid'], ['uuid'])
    op.create_foreign_key('batch_vendorinvoice_fk_vendor',
                          'batch_vendorinvoice', 'vendor',
                          ['vendor_uuid'], ['uuid'])

    # batch_vendorinvoice_row
    op.drop_constraint('vendor_invoice_row_fk_product', 'vendor_invoice_row', type_='foreignkey')
    op.rename_table('vendor_invoice_row', 'batch_vendorinvoice_row')
    op.create_foreign_key('batch_vendorinvoice_row_fk_batch_uuid',
                          'batch_vendorinvoice_row', 'batch_vendorinvoice',
                          ['batch_uuid'], ['uuid'])
    op.create_foreign_key('batch_vendorinvoice_row_fk_product',
                          'batch_vendorinvoice_row', 'product',
                          ['product_uuid'], ['uuid'])
    

def downgrade():

    # remove references to `batch_vendorinvoice`
    op.drop_constraint('batch_vendorinvoice_row_fk_batch_uuid', 'batch_vendorinvoice_row', type_='foreignkey')

    # batch_vendorinvoice
    op.drop_constraint('batch_vendorinvoice_fk_created_by', 'batch_vendorinvoice', type_='foreignkey')
    op.drop_constraint('batch_vendorinvoice_fk_cognized_by', 'batch_vendorinvoice', type_='foreignkey')
    op.drop_constraint('batch_vendorinvoice_fk_executed_by', 'batch_vendorinvoice', type_='foreignkey')
    op.drop_constraint('batch_vendorinvoice_fk_vendor', 'batch_vendorinvoice', type_='foreignkey')
    op.rename_table('batch_vendorinvoice', 'vendor_invoice')
    op.create_foreign_key('vendor_invoice_fk_created_by',
                          'vendor_invoice', 'user',
                          ['created_by_uuid'], ['uuid'])
    op.create_foreign_key('vendor_invoice_fk_cognized_by',
                          'vendor_invoice', 'user',
                          ['cognized_by_uuid'], ['uuid'])
    op.create_foreign_key('vendor_invoice_fk_executed_by',
                          'vendor_invoice', 'user',
                          ['executed_by_uuid'], ['uuid'])
    op.create_foreign_key('vendor_invoice_fk_vendor',
                          'vendor_invoice', 'vendor',
                          ['vendor_uuid'], ['uuid'])

    # batch_vendorinvoice_row
    op.drop_constraint('batch_vendorinvoice_row_fk_product', 'batch_vendorinvoice_row', type_='foreignkey')
    op.rename_table('batch_vendorinvoice_row', 'vendor_invoice_row')
    op.create_foreign_key('vendor_invoice_row_fk_batch',
                          'vendor_invoice_row', 'vendor_invoice',
                          ['batch_uuid'], ['uuid'])
    op.create_foreign_key('vendor_invoice_row_fk_product',
                          'vendor_invoice_row', 'product',
                          ['product_uuid'], ['uuid'])



# -*- coding: utf-8; -*-
"""add pricing_batch.margin_diff

Revision ID: d28d28bc3cd0
Revises: e645676eb37a
Create Date: 2018-11-18 21:02:37.435757

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = 'd28d28bc3cd0'
down_revision = 'e645676eb37a'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # batch_pricing_row
    op.add_column('batch_pricing_row', sa.Column('margin_diff', sa.Numeric(precision=6, scale=3), nullable=True))
    op.add_column('batch_pricing_row', sa.Column('price_diff_percent', sa.Numeric(precision=8, scale=3), nullable=True))


def downgrade():

    # batch_pricing_row
    op.drop_column('batch_pricing_row', 'price_diff_percent')
    op.drop_column('batch_pricing_row', 'margin_diff')



# -*- coding: utf-8; -*-
"""grow permission name

Revision ID: efb7cd318947
Revises: 8adbe48ede2c
Create Date: 2020-08-10 19:33:53.406026

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = 'efb7cd318947'
down_revision = u'8adbe48ede2c'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # permission
    op.alter_column('permission', 'permission', type_=sa.String(length=254))


def downgrade():

    # permission
    op.alter_column('permission', 'permission', type_=sa.String(length=50))



# -*- coding: utf-8; -*-
"""add label_batch.label_code

Revision ID: c5f39e2dec78
Revises: a7b7246e48e2
Create Date: 2018-12-18 19:04:39.624645

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = 'c5f39e2dec78'
down_revision = 'a7b7246e48e2'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # label_batch
    op.add_column('label_batch', sa.Column('label_code', sa.String(length=30), nullable=True))


def downgrade():

    # label_batch
    op.drop_column('label_batch', 'label_code')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2022 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
App Handler
"""

from __future__ import unicode_literals, absolute_import

import os
# import re
import tempfile
import warnings
import logging

import six

from rattail.util import (load_object, load_entry_points,
                          OrderedDict, progress_loop, pretty_quantity)
from rattail.files import temp_path
from rattail.mail import send_email
from rattail.config import parse_list
from rattail.core import get_uuid, Object


log = logging.getLogger(__name__)


class AppHandler(object):
    """
    Base class and default implementation for top-level Rattail app handler.

    aka. "the handler to handle all handlers"

    aka. "one handler to bind them all"
    """
    default_autocompleters = {
        'brands': 'rattail.autocomplete.brands:BrandAutocompleter',
        'customers': 'rattail.autocomplete.customers:CustomerAutocompleter',
        'customers.neworder': 'rattail.autocomplete.customers:CustomerNewOrderAutocompleter',
        'customers.phone': 'rattail.autocomplete.customers:CustomerPhoneAutocompleter',
        'employees': 'rattail.autocomplete.employees:EmployeeAutocompleter',
        'departments': 'rattail.autocomplete.departments:DepartmentAutocompleter',
        'people': 'rattail.autocomplete.people:PersonAutocompleter',
        'people.employees': 'rattail.autocomplete.people:PersonEmployeeAutocompleter',
        'people.neworder': 'rattail.autocomplete.people:PersonNewOrderAutocompleter',
        'products': 'rattail.autocomplete.products:ProductAutocompleter',
        'products.all': 'rattail.autocomplete.products:ProductAllAutocompleter',
        'products.neworder': 'rattail.autocomplete.products:ProductNewOrderAutocompleter',
        'vendors': 'rattail.autocomplete.vendors:VendorAutocompleter',
    }

    def __init__(self, config):
        self.config = config

        # app may not use the db layer, but if so we set the model
        try:
            self.model = config.get_model()
        except: # pragma: no cover
            pass

    def get_title(self, default=None):
        """
        Returns the configured title (name) of the app.

        Default logic invokes
        :meth:`rattail.config.RattailConfig.app_title()` to obtain the
        title.

        :param default: Value to be returned if there is no app title
           configured.

        :returns: Title for the app.
        """
        return self.config.app_title(default=default)

    def get_timezone(self, key='default'):
        """
        Returns a configured time zone.

        Default logic invokes :func:`rattail.time.timezone()` to
        obtain the time zone object.

        :param key: Unique key designating which time zone should be
           returned.  Note that most apps have only one ("default"),
           but may have others defined.
        """
        from rattail.time import timezone
        return timezone(self.config, key)

    def localtime(self, *args, **kwargs):
        """
        Produce or convert a timestamp in the default time zone.

        Default logic invokes :func:`rattail.time.localtime()` to
        obtain the timestamp.  All args and kwargs are passed directly
        to that function.

        :returns: A :class:`python:datetime.datetime` object.  Usually
           this will be timezone-aware but this will depend on the
           args and kwargs you specify.
        """
        from rattail.time import localtime
        return localtime(self.config, *args, **kwargs)

    def make_utc(self, *args, **kwargs):
        """
        Produce or convert a timestamp to UTC time zone.

        Default logic invokes :func:`rattail.time.make_utc()` to
        obtain the timestamp.  All args and kwargs are passed directly
        to that function.

        :returns: A :class:`python:datetime.datetime` object.  Usually
           this will be timezone-naive but this will depend on the
           args and kwargs you specify.
        """
        from rattail.time import make_utc
        return make_utc(*args, **kwargs)

    def load_entry_points(self, group, **kwargs): # pragma: no cover
        warnings.warn("method is deprecated, please use "
                      "util.load_entry_points() function directly",
                      DeprecationWarning)
        return load_entry_points(group, **kwargs)

    def load_object(self, spec):
        """
        Import and/or load and return the object designated by the
        given spec string.

        Default logic invokes :func:`rattail.util.load_object()` to
        obtain the object.

        The syntax of the spec string is not unique to Rattail, but is
        not a universal standard, so deserves a note here.  The spec
        is basically just ``modulepath:objname`` where ``modulepath``
        is the dotted name of the full module where the object
        resides, and ``objname`` is the name of the object within that
        module.  For instance, ``rattail.app:AppHandler`` would be the
        spec for *this* class.

        Note also that the use of the word "object" may be confusing,
        as it does not signify an "instance" but rather an object in
        the generic sense.  Most often a spec will either refer to a
        class or function within the module, although any valid named
        object is possible.

        :param spec: String like ``modulepath:objname`` as described
           above.

        :returns: The object referred to by ``spec``.  If the module
           could not be imported, or did not contain an object of the
           given name, then an error will raise.
        """
        return load_object(spec)

    def next_counter_value(self, session, key, **kwargs):
        """
        Return the next counter value for the given key.

        :param session: Current session for Rattail DB.

        :param key: Unique key indicating the counter for which the
           next value should be fetched.

        :returns: Next value as integer.
        """
        dialect = session.bind.url.get_dialect().name
        if dialect != 'postgresql':
            log.debug("non-postgresql database detected; will use workaround")
            from rattail.db.util import CounterMagic
            magic = CounterMagic(self.config)
            return magic.next_value(session, key)

        # normal (uses postgresql sequence)
        sql = "select nextval('{}_seq')".format(key)
        value = session.execute(sql).scalar()
        return value

    def get_active_stores(self, session, **kwargs):
        """
        Returns the list of "active" stores.  A store is considered
        active if it is *not* marked as archived.

        :param session: Reference to current DB session.

        :returns: Possibly-empty list of
           :class:`~rattail.db.model.stores.Store` records which are
           deemed active.
        """
        import sqlalchemy as sa

        model = self.model
        return session.query(model.Store)\
                      .filter(sa.or_(
                          model.Store.archived == False,
                          model.Store.archived == None))\
                      .order_by(model.Store.id)\
                      .all()

    def get_autocompleter(self, key, **kwargs):
        """
        Returns a new :class:`~rattail.autocomplete.base.Autocompleter`
        instance corresponding to the given key, e.g. ``'products'``.

        The app handler has some hard-coded defaults for the built-in
        autocompleters (see ``default_autocompleters`` in the source
        code).  You can override any of these, and/or add your own
        with custom keys, via config, e.g.:

        .. code-block:: ini

           [rattail]
           autocomplete.products = poser.autocomplete.products:ProductAutocompleter
           autocomplete.otherthings = poser.autocomplete.things:OtherThingAutocompleter

        With the above you can then fetch your custom autocompleter with::

           autocompleter = app.get_autocompleter('otherthings')

        In any case if it can locate the class, it will create an
        instance of it and return that.

        :params key: Unique key for the type of autocompleter you
           need.  Often is a simple string, e.g. ``'customers'`` but
           sometimes there may be a "modifier" with it to get an
           autocompleter with more specific behavior.

           For instance ``'customers.phone'`` would effectively give
           you a customer autocompleter but which searched by phone
           number instead of customer name.

           Note that each key is still a simple string though, and that
           must be "unique" in the sense that only one autocompleter
           can be configured for each key.

        :returns: An :class:`~rattail.autocomplete.base.Autocompleter`
           instance if found, otherwise ``None``.
        """
        spec = self.config.get('rattail', 'autocomplete.{}'.format(key))
        if not spec:
            spec = self.default_autocompleters.get(key)
        if spec:
            return load_object(spec)(self.config)

        raise ValueError("cannot locate autocompleter for key: {}".format(key))

    def get_auth_handler(self, **kwargs):
        """
        Get the configured "auth" handler.

        :returns: The :class:`~rattail.auth.AuthHandler` instance for
           the app.
        """
        if not hasattr(self, 'auth_handler'):
            spec = self.config.get('rattail', 'auth.handler',
                                   default='rattail.auth:AuthHandler')
            factory = load_object(spec)
            self.auth_handler = factory(self.config, **kwargs)
        return self.auth_handler

    def get_batch_handler(self, key, default=None, error=True, **kwargs):
        """
        Get the configured batch handler of the given type.

        :param key: Unique key designating which type of batch handler
           is being requested.

        :param default: Spec string to be used as the default, if no
           handler is configured for the given batch type.  This spec
           string must itself refer to a ``BatchHandler`` class.

        :param error: Flag indicating whether an error should be
           raised if no handler can be found.

        :returns: A :class:`~rattail.batch.handlers.BatchHandler`
           instance of the requested type.  If no such handler can be
           found, and the ``error`` param is false, then ``None`` is
           returned; otherwise an error will raise.
        """
        # spec is assumed to come from config/settings if present,
        # otherwise caller-supplied default is assumed
        spec = self.config.get('rattail.batch', '{}.handler'.format(key),
                               default=default)

        # TODO: this probably should go away?
        # if none of the above gave us a spec, check for common 'importer' type
        if not spec and key == 'importer':
            spec = 'rattail.batch.importer:ImporterBatchHandler'

        if spec:
            Handler = self.load_object(spec)
            return Handler(self.config)

        if error:
            raise ValueError("handler spec not found for batch "
                             "type: {}".format(key))

    def get_board_handler(self, **kwargs):
        """
        Get the configured "board" handler.

        :returns: The :class:`~rattail.board.BoardHandler` instance
           for the app.
        """
        if not hasattr(self, 'board_handler'):
            from rattail.board import get_board_handler
            self.board_handler = get_board_handler(self.config, **kwargs)
        return self.board_handler

    def get_bounce_handler(self, key, **kwargs):
        """
        Get the configured email bounce handler of the given type.

        :param key: Unique key designating which type of bounce
           handler is being requested.

        :returns: A :class:`~rattail.bouncer.handler.BounceHandler`
           instance of the requested type.  If no such handler can be
           found, an error will raise.
        """
        if not hasattr(self, 'bounce_handlers'):
            self.bounce_handlers = {}

        if key not in self.bounce_handlers:
            spec = self.config.get('rattail.bouncer',
                                   '{}.handler'.format(key))
            if not spec and key == 'default':
                spec = 'rattail.bouncer:BounceHandler'
            if not spec:
                raise ValueError("bounce handler spec not found for "
                                 "type: {}".format(key))

            Handler = self.load_object(spec)
            self.bounce_handlers[key] = Handler(self.config, key)

        return self.bounce_handlers[key]

    def get_clientele_handler(self, **kwargs):
        """
        Get the configured "clientele" handler.

        :returns: The :class:`~rattail.clientele.ClienteleHandler`
           instance for the app.
        """
        if not hasattr(self, 'clientele_handler'):
            from rattail.clientele import get_clientele_handler
            self.clientele_handler = get_clientele_handler(self.config, **kwargs)
        return self.clientele_handler

    def get_custorder_handler(self, **kwargs):
        """
        Get the configured "customer order" handler.

        :returns: The :class:`~rattail.custorders.CustomerOrderHandler`
           instance for the app.
        """
        if not hasattr(self, 'custorder_handler'):
            spec = self.config.get('rattail', 'custorders.handler',
                                   default='rattail.custorders:CustomerOrderHandler')
            Handler = self.load_object(spec)
            self.custorder_handler = Handler(self.config)
        return self.custorder_handler

    def get_employment_handler(self, **kwargs):
        """
        Get the configured "employment" handler.

        :returns: The :class:`~rattail.employment.EmploymentHandler`
           instance for the app.
        """
        if not hasattr(self, 'employment_handler'):
            from rattail.employment import get_employment_handler
            self.employment_handler = get_employment_handler(self.config, **kwargs)
        return self.employment_handler

    def get_feature_handler(self, **kwargs):
        """
        Get the configured "feature" handler.

        :returns: The :class:`~rattail.features.handlers.FeatureHandler`
           instance for the app.
        """
        if not hasattr(self, 'feature_handler'):
            from rattail.features import FeatureHandler
            self.feature_handler = FeatureHandler(self.config, **kwargs)
        return self.feature_handler

    def get_email_handler(self, **kwargs):
        """
        Get the configured "email" handler.

        :returns: The :class:`~rattail.mail.EmailHandler` instance for
           the app.
        """
        if not hasattr(self, 'email_handler'):
            from rattail.mail import get_email_handler
            self.email_handler = get_email_handler(self.config, **kwargs)
        return self.email_handler

    def get_mail_handler(self, **kwargs): # pragma: no cover
        warnings.warn("method is deprecated, please use "
                      "AppHandler.get_email_handler() instead",
                      DeprecationWarning)
        return self.get_email_handler(**kwargs)

    def get_all_import_handlers(self, ignore_errors=True, sort=False,
                                **kwargs):
        """
        Returns *all* Import/Export Handler classes which are known to
        exist, i.e.  all which are registered via ``setup.py`` for the
        various packages installed.

        This means it will include both "designated" handlers as well
        as non-designated.  See
        :meth:`get_designated_import_handlers()` if you only want the
        designated ones.

        Note that this will return the *Handler classes* and not
        *handler instances*.

        :param ignore_errors: Normally any errors which come up during
           the loading process are ignored.  Pass ``False`` here to
           force errors to raise, e.g. if you are not seeing the
           results you expect.

        :param sort: If you like the results can be sorted with a
           simple key based on "Source -> Target" labels.

        :returns: List of all registered Import/Export Handler classes.
        """
        # first load all "registered" Handler classes
        Handlers = load_entry_points('rattail.importing',
                                     ignore_errors=ignore_errors)

        # organize registered classes by spec
        specs = {}
        for Handler in six.itervalues(Handlers):
            spec = '{}:{}'.format(Handler.__module__, Handler.__name__)
            specs[spec] = Handler

        # many handlers may not be registered per se, but may be
        # designated via config.  so try to include those too
        for Handler in six.itervalues(Handlers):
            spec = self.get_designated_import_handler_spec(Handler.get_key())
            if spec and spec not in specs:
                specs[spec] = load_object(spec)

        # flatten back to simple list
        Handlers = list(specs.values())

        if sort:
            Handlers.sort(key=lambda h: (h.get_generic_host_title(),
                                         h.get_generic_local_title()))

        return Handlers

    def get_designated_import_handlers(self, with_alternates=False, **kwargs):
        """
        Returns all "designated" import/export handler instances.

        Each "handler type key" can have at most one Handler class
        which is "designated" in the config.  This method collects all
        registered handlers and then sorts out which one is
        designated, for each type key, ultimately returning only the
        designated ones.

        Note that this will return the *handler instances* and not
        *Handler classes*.

        If you have a type key and just need its designated handler,
        see :meth:`get_import_handler()`.

        See also :meth:`get_all_import_handlers()` if you need all
        registered Handler classes.

        :param with_alternates: If you specify ``True`` here then each
           designated handler returned will have an extra attribute
           named ``alternate_handlers``, which will be a list of the
           other "available" (registered) handlers which match the
           designated handler's type key.

           This is probably most / only useful for the Configuration
           UI, to allow admin to change which is designated.

        :returns: List of all designated import/export handler instances.
        """
        grouped = OrderedDict()
        for Handler in self.get_all_import_handlers(**kwargs):
            key = Handler.get_key()
            grouped.setdefault(key, []).append(Handler)

        def find_designated(key, group):
            spec = self.get_designated_import_handler_spec(key)
            if spec:
                for Handler in group:
                    if Handler.get_spec() == spec:
                        return Handler

            if len(group) == 1:
                return group[0]

        designated = []
        for key, group in six.iteritems(grouped):
            Handler = find_designated(key, group)
            if Handler:
                # nb. we must instantiate here b/c otherwise if we
                # assign the `alternate_handlers` attr onto the class,
                # it can affect subclasses as well.  not so with
                # instances though
                handler = Handler(self.config)
                if with_alternates:
                    handler.alternate_handlers = [H for H in group
                                                  if H is not Handler]
                designated.append(handler)

        return designated

    def get_import_handler(self, key, require=False, **kwargs):
        """
        Return the designated import/export handler instance, per the
        given handler type key.

        See also :meth:`get_designated_import_handlers()` if you want
        the full set of designated handlers.

        :param key: A "handler type key", e.g.
           ``'to_rattail.from_rattail.import'``.

        :param require: Specify ``True`` here if you want an error to
           be raised should no handler be found.

        :returns: The import/export handler instance corresponding to
           the given key.  If no handler can be found, then ``None``
           is returned, unless ``require`` param is true, in which
           case error is raised.
        """
        # first try to fetch the handler per designated spec
        spec = self.get_designated_import_handler_spec(key, **kwargs)
        if spec:
            Handler = self.load_object(spec)
            return Handler(self.config)

        # nothing was designated, so leverage logic which already
        # sorts out which handler is "designated" for given key
        designated = self.get_designated_import_handlers()
        for handler in designated:
            if handler.get_key() == key:
                return handler

        if require:
            raise ValueError("Cannot locate handler for key: {}".format(key))

    def get_designated_import_handler(self, *args, **kwargs): # pragma: no cover
        warnings.warn("method is deprecated, please use "
                      "AppHandler.get_import_handler() instead",
                      DeprecationWarning)
        return self.get_import_handler(*args, **kwargs)

    def get_designated_import_handler_spec(self, key, require=False, **kwargs):
        """
        Return the designated import handler "spec" string for the
        given type key.

        :param key: Unique key indicating the type of import handler.

        :require: Flag indicating whether an error should be raised if no
           handler is found.

        :returns: Spec string for the designated handler.  If none is
           found, then ``None`` is returned *unless* the ``require``
           param is true, in which case an error is raised.
        """
        spec = self.config.get('rattail.importing',
                               '{}.handler'.format(key))
        if spec:
            return spec

        legacy_setting = self.config.get('rattail.importing',
                                         '{}.legacy_handler_setting'.format(key))
        if legacy_setting:
            legacy_setting = parse_list(legacy_setting)
            if len(legacy_setting) == 2:
                section, option = legacy_setting
                spec = self.config.get(section, option)
                if spec:
                    return spec

        spec = self.config.get('rattail.importing',
                               '{}.default_handler'.format(key))
        if spec:
            return spec

        if require:
            raise ValueError("Cannot locate handler spec for key: {}".format(key))

    def get_label_handler(self, **kwargs):
        """
        Get the configured "label" handler.

        See also :doc:`rattail-manual:base/handlers/other/labels`.

        :returns: The :class:`~rattail.labels.LabelHandler` instance
           for the app.
        """
        if not hasattr(self, 'label_handler'):
            spec = self.config.get('rattail', 'labels.handler',
                                   default='rattail.labels:LabelHandler')
            factory = self.load_object(spec)
            self.label_handler = factory(self.config, **kwargs)
        return self.label_handler

    def get_membership_handler(self, **kwargs):
        """
        Get the configured "membership" handler.

        See also :doc:`rattail-manual:base/handlers/other/membership`.

        :returns: The :class:`~rattail.membership.MembershipHandler`
           instance for the app.
        """
        if not hasattr(self, 'membership_handler'):
            spec = self.config.get('rattail', 'membership.handler',
                                   default='rattail.membership:MembershipHandler')
            factory = load_object(spec)
            self.membership_handler = factory(self.config, **kwargs)
        return self.membership_handler

    def get_people_handler(self, **kwargs):
        """
        Get the configured "people" handler.

        See also :doc:`rattail-manual:base/handlers/other/people`.

        :returns: The :class:`~rattail.people.PeopleHandler` instance
           for the app.
        """
        if not hasattr(self, 'people_handler'):
            spec = self.config.get('rattail', 'people.handler',
                                   default='rattail.people:PeopleHandler')
            factory = load_object(spec)
            self.people_handler = factory(self.config, **kwargs)
        return self.people_handler

    def get_products_handler(self, **kwargs):
        """
        Get the configured "products" handler.

        :returns: The :class:`~rattail.products.ProductsHandler`
           instance for the app.
        """
        if not hasattr(self, 'products_handler'):
            from rattail.products import get_products_handler
            self.products_handler = get_products_handler(self.config, **kwargs)
        return self.products_handler

    def get_report_handler(self, **kwargs):
        """
        Get the configured "reports" handler.

        :returns: The :class:`~rattail.reporting.handlers.ReportHandler`
           instance for the app.
        """
        if not hasattr(self, 'report_handler'):
            from rattail.reporting import get_report_handler
            self.report_handler = get_report_handler(self.config, **kwargs)
        return self.report_handler

    def get_problem_report_handler(self, **kwargs):
        """
        Get the configured "problem reports" handler.

        :returns: The :class:`~rattail.problems.handlers.ProblemReportHandler`
           instance for the app.
        """
        if not hasattr(self, 'problem_report_handler'):
            from rattail.problems import get_problem_report_handler
            self.problem_report_handler = get_problem_report_handler(
                self.config, **kwargs)
        return self.problem_report_handler

    def get_trainwreck_handler(self, **kwargs):
        """
        Get the configured "trainwreck" handler.

        :returns: The :class:`~rattail.trainwreck.handler.TrainwreckHandler`
           instance for the app.
        """
        if not hasattr(self, 'trainwreck_handler'):
            spec = self.config.get('trainwreck', 'handler',
                                   default='rattail.trainwreck.handler:TrainwreckHandler')
            Handler = self.load_object(spec)
            self.trainwreck_handler = Handler(self.config)
        return self.trainwreck_handler

    def get_vendor_handler(self, **kwargs):
        """
        Get the configured "vendor" handler.

        :returns: The :class:`~rattail.vendors.handler.VendorHandler`
           instance for the app.
        """
        if not hasattr(self, 'vendor_handler'):
            spec = self.config.get('rattail', 'vendors.handler',
                                   default='rattail.vendors:VendorHandler')
            factory = self.load_object(spec)
            self.vendor_handler = factory(self.config, **kwargs)
        return self.vendor_handler

    def progress_loop(self, *args, **kwargs):
        """
        Run a given function for a given sequence, and optionally show
        a progress indicator.

        Default logic invokes the :func:`rattail.util.progress_loop()`
        function; see that for more details.
        """
        return progress_loop(*args, **kwargs)

    def make_object(self, **kwargs):
        """
        Create and return a generic object.  All kwargs will be
        assigned as attributes to the object.
        """
        return Object(**kwargs)

    def get_session(self, obj):
        """
        Returns the SQLAlchemy session with which the given object is
        associated.  Simple convenience wrapper around
        :func:`sqlalchemy:sqlalchemy.orm.object_session()`.
        """
        from sqlalchemy import orm

        return orm.object_session(obj)

    def make_session(self, user=None, **kwargs):
        """
        Creates and returns a new SQLAlchemy session for the Rattail DB.

        :param user: A "user-ish" object which should be considered
           responsible for changes made during the session.  Can be
           either a :class:`~rattail.db.model.users.User` object, or
           just a (string) username.  If none is specified then the
           config will be consulted for a default.

        :returns: A :class:`rattail.db.Session` instance.
        """
        from rattail.db import Session

        # always try to set default continuum user if possible
        if 'continuum_user' not in kwargs:
            if not user:
                user = self.config.get('rattail', 'runas.default')
            if user:
                kwargs['continuum_user'] = user

        return Session(**kwargs)

    def cache_model(self, session, model, **kwargs):
        """
        Convenience method which invokes
        :func:`rattail.db.cache.cache_model()` with the given model
        and keyword arguments.
        """
        from rattail.db import cache
        return cache.cache_model(session, model, **kwargs)

    def make_temp_dir(self, **kwargs):
        """
        Create a temporary directory.  This is mostly a convenience
        wrapper around the built-in :func:`python:tempfile.mkdtemp()`.
        However by default it will attempt to place the temp folder
        underneath the configured "workdir", e.g.:

        .. code-block:: ini

           [rattail]
           workdir = /srv/envs/poser/app/work
        """
        if 'dir' not in kwargs:
            workdir = self.config.workdir(require=False)
            if workdir:
                tmpdir = os.path.join(workdir, 'tmp')
                if not os.path.exists(tmpdir):
                    os.makedirs(tmpdir)
                kwargs['dir'] = tmpdir
        return tempfile.mkdtemp(**kwargs)

    def make_temp_file(self, **kwargs):
        """
        Reserve a temporary filename.  This is mostly a convenience
        wrapper around the built-in :func:`python:tempfile.mkstemp()`.
        However by default it will attempt to place the temp file
        underneath the configured "workdir", e.g.:

        .. code-block:: ini

           [rattail]
           workdir = /srv/envs/poser/app/work
        """
        if 'dir' not in kwargs:
            workdir = self.config.workdir(require=False)
            if workdir:
                tmpdir = os.path.join(workdir, 'tmp')
                if not os.path.exists(tmpdir):
                    os.makedirs(tmpdir)
                kwargs['dir'] = tmpdir
        return temp_path(**kwargs)

    def make_uuid(self):
        """
        Generate a new UUID value.

        :returns: UUID value as 32-character string.
        """
        return get_uuid()

    def normalize_phone_number(self, number, **kwargs):
        """
        Normalize the given phone number, to a "common" format that
        can be more easily worked with for sync logic etc.  In
        practice this usually just means stripping all non-digit
        characters from the string.  The idea is that phone number
        data from any system can be "normalized" and thereby compared
        directly to see if they differ etc.

        Default logic will invoke
        :func:`rattail.db.util.normalize_phone_number()`.

        :param number: Raw phone number string e.g. as found in some
           data source.

        :returns: Normalized string.
        """
        from rattail.db.util import normalize_phone_number

        return normalize_phone_number(number)

    def phone_number_is_invalid(self, number):
        """
        This method should validate the given phone number string, and
        if the number is *not* considered valid, this method should
        return the reason.

        :param number: Raw phone number string e.g. as found in some
           data source.

        :returns: String describing reason the number is invalid, or
           ``None`` if the number is deemed valid.
        """
        # strip non-numeric chars, and make sure we have 10 left
        normal = self.normalize_phone_number(number)
        if len(normal) != 10:
            return "Phone number must have 10 digits"

    def format_phone_number(self, number):
        """
        Returns a "properly formatted" string based on the given phone
        number.

        Default logic invokes
        :func:`rattail.db.util.format_phone_number()`.

        :param number: Raw phone number string e.g. as found in some
           data source.

        :returns: Formatted phone number string.
        """
        from rattail.db.util import format_phone_number

        return format_phone_number(number)

    def make_gpc(self, value, **kwargs):
        """
        Make and return a :class:`~rattail.gpc.GPC` instance from the
        given value.

        Default logic will invoke
        :meth:`~rattail.products.ProductsHandler.make_gpc()` of the
        products handler; see also :meth:`get_products_handler()`.
        """
        products_handler = self.get_products_handler()
        return products_handler.make_gpc(value, **kwargs)

    def render_gpc(self, value, **kwargs):
        """
        Returns a human-friendly display string for the given GPC
        value.

        :param value: A :class:`~rattail.gpc.GPC` instance.

        :returns: Display string for the GPC, or ``None`` if the value
           provided is not a GPC.
        """
        if value:
            return value.pretty()

    def render_upc(self, value, **kwargs): # pragma: no cover
        warnings.warn("method is deprecated, please use "
                      "render_gpc() method instead",
                      DeprecationWarning)
        return self.render_gpc(value, **kwargs)

    def render_currency(self, value, scale=2, **kwargs):
        """
        Must return a human-friendly display string for the given
        currency value, e.g. ``Decimal('4.20')`` becomes ``"$4.20"``.

        :param value: Either a :class:`python:decimal.Decimal` or
           :class:`python:float` value.

        :param scale: Number of decimal digits to be displayed.

        :returns: Display string for the value.
        """
        if value is not None:
            if value < 0:
                fmt = "(${{:0,.{}f}})".format(scale)
                return fmt.format(0 - value)
            fmt = "${{:0,.{}f}}".format(scale)
            return fmt.format(value)

    def render_quantity(self, value, **kwargs):
        """
        Return a human-friendly display string for the given quantity
        value, e.g. ``1.000`` becomes ``"1"``.

        :param value: The quantity to be rendered.

        :returns: Display string for the quantity.
        """
        return pretty_quantity(value, **kwargs)

    def render_cases_units(self, cases, units):
        """
        Render a human-friendly string showing the given number of
        cases and/or units.  For instance::

           >>> app.render_cases_units(1, None)
           '1 case'

           >>> app.render_cases_units(None, 1)
           '1 unit'

           >>> app.render_cases_units(3, 2)
           '3 cases + 2 units'

        :param cases: Number of cases (can be zero or ``None``).

        :param units: Number of units (can be zero or ``None``).

        :returns: Display string for the given values.
        """
        if cases is not None:
            label = "case" if abs(cases) == 1 else "cases"
            cases = "{} {}".format(self.render_quantity(cases), label)

        if units is not None:
            label = "unit" if abs(units) == 1 else "units"
            units = "{} {}".format(self.render_quantity(units), label)

        if cases and units:
            return "{} + {}".format(cases, units)

        return cases or units

    def render_date(self, value, **kwargs):
        """
        Return a human-friendly display string for the given date.

        :param value: A :class:`python:datetime.date` instance.

        :returns: Display string for the date.
        """
        if value is not None:
            return value.strftime('%Y-%m-%d')

    def render_datetime(self, value, **kwargs):
        """
        Return a human-friendly display string for the given datetime.
        """
        if value is not None:
            return value.strftime('%Y-%m-%d %I:%M:%S %p')

    def send_email(self, key, data={}, **kwargs):
        """
        Send an email message of the given type.

        See :func:`rattail.mail.send_email()` for more info.
        """
        send_email(self.config, key, data, **kwargs)


class GenericHandler(object):
    """
    Base class for misc. "generic" feature handlers.

    Most handlers which exist for sake of business logic, should inherit from
    this.
    """

    def __init__(self, config, **kwargs):
        self.config = config
        self.enum = self.config.get_enum()
        self.model = self.config.get_model()
        self.app = self.config.get_app()

    def progress_loop(self, *args, **kwargs): # pragma: no cover
        warnings.warn("method is deprecated, please use "
                      "AppHandler.progress_loop() method instead",
                      DeprecationWarning)
        return self.app.progress_loop(*args, **kwargs)

    def get_session(self, obj): # pragma: no cover
        warnings.warn("method is deprecated, please use "
                      "AppHandler.get_session() method instead",
                      DeprecationWarning)
        return self.app.get_session(obj)

    def make_session(self): # pragma: no cover
        warnings.warn("method is deprecated, please use "
                      "AppHandler.make_session() method instead",
                      DeprecationWarning)
        return self.app.make_session()

    def cache_model(self, session, model, **kwargs): # pragma: no cover
        warnings.warn("method is deprecated, please use "
                      "AppHandler.cache_model() method instead",
                      DeprecationWarning)
        return self.app.cache_model(session, model, **kwargs)


def make_app(config, **kwargs): # pragma: no cover
    warnings.warn("function is deprecated, please use "
                  "RattailConfig.get_app() method instead",
                  DeprecationWarning)
    return config.get_app()



# -*- coding: utf-8 -*-
"""add employee.full_time

Revision ID: 86632a7bc5b5
Revises: b9658d212053
Create Date: 2017-05-16 10:39:52.906965

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '86632a7bc5b5'
down_revision = u'b9658d212053'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


def upgrade():

    # employee
    op.add_column('employee', sa.Column('full_time', sa.Boolean(), nullable=True))
    op.add_column('employee', sa.Column('full_time_start', sa.Date(), nullable=True))


def downgrade():

    # employee
    # op.drop_column('employee', 'full_time_start')
    op.drop_column('employee', 'full_time')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Mail Monitor Configuration
"""

from __future__ import unicode_literals, absolute_import

import six

from rattail.config import ConfigProfile, parse_list
from rattail.exceptions import ConfigurationError


class MailMonitorAccountProfile(ConfigProfile):
    """
    Simple class to hold configuration specific to an "account" within
    the MailMon setup.
    """
    section = 'rattail.mailmon'

    def load(self):
        self.server = self._config_string('server')
        self.username = self._config_string('username')
        self.password = self._config_string('password')
        self.error_delay = self._config_int('error_delay',
                                            minimum=0,
                                            default=60)
        self.recycle_delay = self._config_int('recycle_delay',
                                              minimum=0)


class MailMonitorProfile(ConfigProfile):
    """
    Simple class to hold configuration for a MailMon "profile".  Each
    profile determines which email folder(s) will be watched for new
    messages, and which action(s) will then be invoked to process the
    messages.
    """
    section = 'rattail.mailmon'

    def load(self):

        self.imap_account = self._config_string('imap.account')
        self.imap_folder = self._config_string('imap.folder')
        self.imap_unread_only = self._config_boolean('imap.unread_only')
        self.imap_delay = self._config_int('imap.delay', default=120)
        self.imap_recycle = self._config_int('imap.recycle', default=0,
                                             minimum=0)

        self.max_batch_size = self._config_int('max_batch_size', default=100)

        self.load_defaults()
        self.load_actions()

    def validate(self):
        """
        Validate the configuration for current profile.
        """
        if not self.actions:
            raise ConfigurationError("mailmon profile '{}' has no valid "
                                     "actions to invoke".format(self.key))


def load_mailmon_profiles(config):
    """
    Load all active mail monitor profiles defined within configuration.
    """
    # make sure we have our top-level directives
    monitor_keys = config.get('rattail.mailmon', 'monitor')
    if not monitor_keys:
        raise ConfigurationError(
            "The mail monitor configuration does not specify any profiles "
            "to be monitored.  Please defined the 'monitor' option within "
            "the [rattail.mailmon] section of your config file.")
    account_keys = config.get('rattail.mailmon', 'accounts')
    if not account_keys:
        raise ConfigurationError(
            "The mail monitor configuration does not specify any accounts "
            "to which to connect.  Please defined the 'accounts' option within "
            "the [rattail.mailmon] section of your config file.")

    # first load the accounts
    accounts = {}
    for key in parse_list(account_keys):
        profile = MailMonitorAccountProfile(config, key,
                                            prefix='account.{}'.format(key))
        accounts[key] = profile

    # now load the monitor profiles
    monitored = {'__accounts__': accounts}
    for key in parse_list(monitor_keys):
        profile = MailMonitorProfile(config, key)

        # only monitor this profile if it uses valid account
        if profile.imap_account not in accounts:
            log.warning("profile references invalid account (%s): %s",
                        profile.imap_account, profile)

        else:
            # only monitor this profile if it validates
            try:
                profile.validate()
            except ConfigurationError as error:
                log.warning(six.text_type(error))
            else:
                monitored[key] = profile

    # finally let each account know what it should monitor
    for account in six.itervalues(accounts):
        account.monitored = [m for k, m in monitored.items()
                             if k != '__accounts__'
                             and m.imap_account == account.key]

    return monitored



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data Models for Customers
"""

from __future__ import unicode_literals, absolute_import

import datetime

import six
import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.ext.orderinglist import ordering_list

from rattail.db.model import (Base, uuid_column, getset_factory,
                              PhoneNumber, EmailAddress, MailingAddress,
                              Person, Note, User)
from .contact import ContactMixin


class Customer(ContactMixin, Base):
    """
    Represents a customer account.

    Customer accounts may consist of more than one person, in some cases.
    """
    __tablename__ = 'customer'
    __versioned__ = {}

    uuid = uuid_column()

    id = sa.Column(sa.String(length=20), nullable=True, doc="""
    String ID for the customer, if known/relevant.  This may or may not
    correspond to the :attr:`number`, depending on your system.
    """)

    number = sa.Column(sa.Integer(), nullable=True, doc="""
    Customer number, if known/relevant.  This may or may not correspond to the
    :attr:`id`, depending on your system.
    """)

    name = sa.Column(sa.String(length=255))
    email_preference = sa.Column(sa.Integer())

    wholesale = sa.Column(sa.Boolean(), nullable=True, doc="""
    Flag indicating whether the customer is a "wholesale" account - whatever
    that happens to mean for your business logic.
    """)

    active_in_pos = sa.Column(sa.Boolean(), nullable=True, doc="""
    Whether or not the customer account should be "active" within the POS
    system, if applicable.  Whether/how this field is populated and/or
    leveraged are up to your system.
    """)

    active_in_pos_sticky = sa.Column(sa.Boolean(), nullable=False, default=False, doc="""
    Whether or not the customer account should *always* be "active" within the
    POS system.  This field may be useful if :attr:`active_in_pos` gets set
    dynamically.
    """)

    invalid_address = sa.Column(sa.Boolean(), nullable=True, doc="""
    Flag indicating the customer's mailing address(es) on file are invalid.
    """)

    def __str__(self):
        if six.PY2:
            return unicode(self).encode('utf8')
        return self.name or ""

    if six.PY2:
        def __unicode__(self):
            return self.name or ""

    def add_email_address(self, address, type='Home'):
        email = CustomerEmailAddress(address=address, type=type)
        self.emails.append(email)
        return email

    def add_phone_number(self, number, type='Home'):
        phone = CustomerPhoneNumber(number=number, type=type)
        self.phones.append(phone)
        return phone

    def add_mailing_address(self, **kwargs):
        addr = CustomerMailingAddress(**kwargs)
        self.addresses.append(addr)
        return addr

    @property
    def employee(self):
        """
        Return the employee associated with the customer, if any.  Assumes a
        certain "typical" relationship path.
        """
        if self.person:
            return self.person.employee

    def first_person(self):
        """
        Convenience method to retrieve the "first" Person record which is
        associated with this customer, or ``None``.
        """
        if self.people:
            return self.people[0]

    def only_person(self, require=True):
        """
        Convenience method to retrieve the one and only Person record which is
        associated with this customer.  An error will be raised if there is not
        exactly one person associated.
        """
        if len(self.people) > 1 or (require and not self.people):
            raise ValueError("customer {} should have 1 person but instead has {}: {}".format(
                self.uuid, len(self.people), self))
        return self.people[0] if self.people else None

    def only_member(self, require=True):
        """
        Convenience method to retrieve the one and only Member record which is
        associated with this customer.  If ``require=True`` then an error will
        be raised if there is not exactly one member found.
        """
        if len(self.members) > 1 or (require and not self.members):
            raise ValueError("customer {} should have 1 member but instead has {}: {}".format(
                self.uuid, len(self.members), self))
        return self.members[0] if self.members else None


class CustomerPhoneNumber(PhoneNumber):
    """
    Represents a phone (or fax) number associated with a :class:`Customer`.
    """

    __mapper_args__ = {'polymorphic_identity': 'Customer'}


Customer._contact_phone_model = CustomerPhoneNumber

Customer.phones = orm.relationship(
    CustomerPhoneNumber,
    backref='customer',
    primaryjoin=CustomerPhoneNumber.parent_uuid == Customer.uuid,
    foreign_keys=[CustomerPhoneNumber.parent_uuid],
    collection_class=ordering_list('preference', count_from=1),
    order_by=CustomerPhoneNumber.preference,
    cascade='save-update, merge, delete, delete-orphan')

Customer.phone = orm.relationship(
    CustomerPhoneNumber,
    primaryjoin=sa.and_(
        CustomerPhoneNumber.parent_uuid == Customer.uuid,
        CustomerPhoneNumber.preference == 1),
    foreign_keys=[CustomerPhoneNumber.parent_uuid],
    uselist=False,
    viewonly=True)


class CustomerEmailAddress(EmailAddress):
    """
    Represents an email address associated with a :class:`Customer`.
    """

    __mapper_args__ = {'polymorphic_identity': 'Customer'}


Customer._contact_email_model = CustomerEmailAddress

Customer.emails = orm.relationship(
    CustomerEmailAddress,
    backref='customer',
    primaryjoin=CustomerEmailAddress.parent_uuid == Customer.uuid,
    foreign_keys=[CustomerEmailAddress.parent_uuid],
    collection_class=ordering_list('preference', count_from=1),
    order_by=CustomerEmailAddress.preference,
    cascade='save-update, merge, delete, delete-orphan')

Customer.email = orm.relationship(
    CustomerEmailAddress,
    primaryjoin=sa.and_(
        CustomerEmailAddress.parent_uuid == Customer.uuid,
        CustomerEmailAddress.preference == 1),
    foreign_keys=[CustomerEmailAddress.parent_uuid],
    uselist=False,
    viewonly=True)


class CustomerMailingAddress(MailingAddress):
    """
    Represents a mailing address for a customer
    """
    __mapper_args__ = {'polymorphic_identity': 'Customer'}


Customer._contact_address_model = CustomerMailingAddress

Customer.addresses = orm.relationship(
    CustomerMailingAddress,
    backref='customer',
    primaryjoin=CustomerMailingAddress.parent_uuid == Customer.uuid,
    foreign_keys=[CustomerMailingAddress.parent_uuid],
    collection_class=ordering_list('preference', count_from=1),
    order_by=CustomerMailingAddress.preference,
    cascade='all, delete-orphan')

Customer.address = orm.relationship(
    CustomerMailingAddress,
    primaryjoin=sa.and_(
        CustomerMailingAddress.parent_uuid == Customer.uuid,
        CustomerMailingAddress.preference == 1),
    foreign_keys=[CustomerMailingAddress.parent_uuid],
    uselist=False,
    viewonly=True)


class CustomerNote(Note):
    """
    Represents a note attached to a customer.
    """
    __mapper_args__ = {'polymorphic_identity': 'Customer'}

    customer = orm.relationship(
        Customer,
        primaryjoin='Customer.uuid == CustomerNote.parent_uuid',
        foreign_keys='CustomerNote.parent_uuid',
        doc="""
        Reference to the customer to which this note is attached.
        """,
        backref=orm.backref(
            'notes',
            primaryjoin='CustomerNote.parent_uuid == Customer.uuid',
            foreign_keys='CustomerNote.parent_uuid',
            order_by='CustomerNote.created',
            cascade='all, delete-orphan',
            doc="""
            Sequence of notes which belong to the customer.
            """))


@six.python_2_unicode_compatible
class CustomerGroup(Base):
    """
    Represents an arbitrary group to which customers may belong.
    """
    __tablename__ = 'customer_group'
    __versioned__ = {}

    uuid = uuid_column()
    id = sa.Column(sa.String(length=20))
    name = sa.Column(sa.String(length=255))

    def __str__(self):
        return self.name or ''


class CustomerGroupAssignment(Base):
    """
    Represents the assignment of a customer to a group.
    """
    __tablename__ = 'customer_x_group'
    __table_args__ = (
        sa.ForeignKeyConstraint(['group_uuid'], ['customer_group.uuid'], name='customer_x_group_fk_group'),
        sa.ForeignKeyConstraint(['customer_uuid'], ['customer.uuid'], name='customer_x_group_fk_customer'),
    )
    __versioned__ = {}

    uuid = uuid_column()
    customer_uuid = sa.Column(sa.String(length=32), nullable=False)
    group_uuid = sa.Column(sa.String(length=32), nullable=False)
    ordinal = sa.Column(sa.Integer(), nullable=False)

    group = orm.relationship(CustomerGroup)


Customer._groups = orm.relationship(
    CustomerGroupAssignment, backref='customer',
    collection_class=ordering_list('ordinal', count_from=1),
    order_by=CustomerGroupAssignment.ordinal,
    cascade='save-update, merge, delete, delete-orphan')

Customer.groups = association_proxy(
    '_groups', 'group',
    getset_factory=getset_factory,
    creator=lambda g: CustomerGroupAssignment(group=g))


class CustomerPerson(Base):
    """
    Represents the association between a person and a customer account.
    """
    __tablename__ = 'customer_x_person'
    __table_args__ = (
        sa.ForeignKeyConstraint(['customer_uuid'], ['customer.uuid'], name='customer_x_person_fk_customer'),
        sa.ForeignKeyConstraint(['person_uuid'], ['person.uuid'], name='customer_x_person_fk_person'),
        sa.Index('customer_x_person_ix_customer', 'customer_uuid'),
        sa.Index('customer_x_person_ix_person', 'person_uuid'),
    )
    __versioned__ = {}

    uuid = uuid_column()
    customer_uuid = sa.Column(sa.String(length=32), nullable=False)
    person_uuid = sa.Column(sa.String(length=32), nullable=False)
    ordinal = sa.Column(sa.Integer(), nullable=False)

    customer = orm.relationship(Customer, back_populates='_people')

    person = orm.relationship(Person)


Customer._people = orm.relationship(
    CustomerPerson, back_populates='customer',
    primaryjoin=CustomerPerson.customer_uuid == Customer.uuid,
    collection_class=ordering_list('ordinal', count_from=1),
    order_by=CustomerPerson.ordinal,
    cascade='save-update, merge, delete, delete-orphan')

Customer.people = association_proxy(
    '_people', 'person',
    getset_factory=getset_factory,
    creator=lambda p: CustomerPerson(person=p))

Customer._person = orm.relationship(
    CustomerPerson,
    primaryjoin=sa.and_(
        CustomerPerson.customer_uuid == Customer.uuid,
        CustomerPerson.ordinal == 1),
    uselist=False,
    viewonly=True)

Customer.person = association_proxy(
    '_person', 'person',
    getset_factory=getset_factory)

Person._customers = orm.relationship(
    CustomerPerson,
    primaryjoin=CustomerPerson.person_uuid == Person.uuid,
    viewonly=True)

Person.customers = association_proxy('_customers', 'customer',
                                     getset_factory=getset_factory,
                                     creator=lambda c: CustomerPerson(customer=c))


class PendingCustomer(Base):
    """
    A "pending" customer record, used for new customer entry workflow.
    """
    __tablename__ = 'pending_customer'
    __table_args__ = (
        sa.ForeignKeyConstraint(['user_uuid'], ['user.uuid'], name='pending_customer_fk_user'),
    )

    uuid = uuid_column()

    user_uuid = sa.Column(sa.String(length=32), nullable=False)
    user = orm.relationship(
        User,
        doc="""
        Referencef to the :class:`~rattail:rattail.db.model.User` who
        first entered the record.
        """)

    created = sa.Column(sa.DateTime(), nullable=False, default=datetime.datetime.utcnow, doc="""
    Timestamp when the record was first created.
    """)

    # Customer fields
    id = sa.Column(sa.String(length=20), nullable=True)

    # Person fields
    first_name = sa.Column(sa.String(length=50), nullable=True)
    middle_name = sa.Column(sa.String(length=50), nullable=True)
    last_name = sa.Column(sa.String(length=50), nullable=True)
    display_name = sa.Column(sa.String(length=100), nullable=True)

    # Phone fields
    phone_number = sa.Column(sa.String(length=20), nullable=True)
    phone_type = sa.Column(sa.String(length=15), nullable=True)

    # Email fields
    email_address = sa.Column(sa.String(length=255), nullable=True)
    email_type = sa.Column(sa.String(length=15), nullable=True)

    # Address fields
    address_street = sa.Column(sa.String(length=100), nullable=True)
    address_street2 = sa.Column(sa.String(length=100), nullable=True)
    address_city = sa.Column(sa.String(length=60), nullable=True)
    address_state = sa.Column(sa.String(length=2), nullable=True)
    address_zipcode = sa.Column(sa.String(length=10), nullable=True)
    address_type = sa.Column(sa.String(length=15), nullable=True)

    # workflow fields
    status_code = sa.Column(sa.Integer(), nullable=True, doc="""
    Status indicator for the new customer record.
    """)

    def __str__(self):
        if six.PY2:
            return (self.display_name or "").encode('utf_8')
        return self.display_name or ""

    if six.PY2:
        def __unicode__(self):
            return self.display_name or ""



# -*- coding: utf-8 -*-
"""add report_output

Revision ID: 0a5a866f2a3d
Revises: 6ef6db00abe8
Create Date: 2017-03-21 21:57:03.435177

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '0a5a866f2a3d'
down_revision = u'6ef6db00abe8'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


def upgrade():

    # report_output
    op.create_table('report_output',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('created', sa.DateTime(), nullable=False),
                    sa.Column('created_by_uuid', sa.String(length=32), nullable=False),
                    sa.Column('record_count', sa.Integer(), nullable=True),
                    sa.Column('report_name', sa.String(length=255), nullable=True),
                    sa.Column('filename', sa.String(length=255), nullable=True),
                    sa.ForeignKeyConstraint(['created_by_uuid'], [u'user.uuid'], name=u'report_output_fk_created_by'),
                    sa.PrimaryKeyConstraint('uuid')
    )


def downgrade():

    # report_output
    op.drop_table('report_output')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2022 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data model for label batches
"""

from __future__ import unicode_literals, absolute_import

import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.ext.orderinglist import ordering_list

from rattail.db.model import Base, LabelProfile, BatchMixin, ProductBatchRowMixin, getset_factory
from rattail.db.core import uuid_column, filename_column


class LabelBatch(BatchMixin, Base):
    """
    Primary data model for label batches.
    """
    __tablename__ = 'batch_labels'
    __batchrow_class__ = 'LabelBatchRow'
    batch_key = 'labels'
    model_title = "Label Batch"
    model_title_plural = "Label Batches"

    @declared_attr
    def __table_args__(cls):
        return cls.__default_table_args__() + (
            sa.ForeignKeyConstraint(['handheld_batch_uuid'], ['batch_handheld.uuid'],
                                    name='batch_labels_fk_handheld_batch'),
            sa.ForeignKeyConstraint(['label_profile_uuid'], ['label_profile.uuid'],
                                    name='batch_labels_fk_label_profile'),
        )

    handheld_batch_uuid = sa.Column(sa.String(length=32), nullable=True)

    handheld_batch = orm.relationship(
        'HandheldBatch',
        doc="""
        Reference to the handheld batch from which this label batch originated.
        """,
        backref=orm.backref(
            'label_batch',
            uselist=False,
            doc="""
            Reference to the label batch to which this handheld batch was converted.
            """))

    filename = filename_column()

    static_prices = sa.Column(sa.Boolean(), nullable=True, default=False, doc="""
    Flag indicating whether product prices within the batch should be
    considered "static" or if they should be updated during batch refresh,
    i.e. read from current product record.
    """)

    label_code = sa.Column(sa.String(length=30), nullable=True, doc="""
    Code (string) of the label type (profile) which should be used when
    printing the rows.  Note that each row may override this.
    """)

    label_profile_uuid = sa.Column(sa.String(length=32), nullable=True)
    label_profile = orm.relationship(
        LabelProfile,
        doc="""
        Reference to the label "profile" which contains details of how the
        rows should be printed.  Note that each row may override this.
        """)


class LabelBatchFromHandheld(Base):
    """
    Represents association betwewn a handheld and label batch.
    """
    __tablename__ = 'batch_labels_handheld'
    __table_args__ = (
        sa.ForeignKeyConstraint(['batch_uuid'], ['batch_labels.uuid'],
                                name='batch_labels_handheld_fk_batch'),
        sa.ForeignKeyConstraint(['handheld_uuid'], ['batch_handheld.uuid'],
                                name='batch_labels_handheld_fk_handheld'),
    )

    uuid = uuid_column()

    batch_uuid = sa.Column(sa.String(length=32), nullable=False)
    ordinal = sa.Column(sa.Integer(), nullable=False)
    batch = orm.relationship(
        LabelBatch,
        doc="""
        Reference to the label batch, with which handheld batch(es) are associated.
        """,
        backref=orm.backref(
            '_handhelds',
            collection_class=ordering_list('ordinal', count_from=1),
            order_by=ordinal,
            cascade='all, delete-orphan',
            doc="""
            Sequence of raw label / handheld batch associations.
            """))
    
    handheld_uuid = sa.Column(sa.String(length=32), nullable=False)
    handheld = orm.relationship(
        'HandheldBatch',
        doc="""
        Reference to the handheld batch from which this label batch originated.
        """,
        backref=orm.backref(
            '_label_batch',
            uselist=False,
            cascade='all, delete-orphan',
            doc="""
            Indirect reference to the label batch to which this handheld batch was converted.
            """))


LabelBatch.handheld_batches = association_proxy('_handhelds', 'handheld',
                                                creator=lambda batch: LabelBatchFromHandheld(handheld=batch),
                                                getset_factory=getset_factory)


class LabelBatchRow(ProductBatchRowMixin, Base):
    """
    Row of data within a label batch.
    """
    __tablename__ = 'batch_labels_row'
    __batch_class__ = LabelBatch

    @declared_attr
    def __table_args__(cls):
        return cls.__default_table_args__() + (
            sa.ForeignKeyConstraint(['label_profile_uuid'], ['label_profile.uuid'],
                                    name='batch_labels_row_fk_label_profile'),
        )

    STATUS_OK                           = 1
    STATUS_PRODUCT_NOT_FOUND            = 2
    STATUS_LABEL_PROFILE_NOT_FOUND      = 3
    STATUS_REGULAR_PRICE_UNKNOWN        = 4
    STATUS_PRODUCT_APPEARS_INACTIVE     = 5

    STATUS = {
        STATUS_OK                       : "ok",
        STATUS_PRODUCT_NOT_FOUND        : "product not found",
        STATUS_LABEL_PROFILE_NOT_FOUND  : "label profile not found",
        STATUS_REGULAR_PRICE_UNKNOWN    : "regular price unknown",
        STATUS_PRODUCT_APPEARS_INACTIVE : "product appears inactive",
    }

    ####################
    # label fields
    ####################

    label_code = sa.Column(sa.String(length=30), nullable=True, doc="""
    Code (string) of the label type (profile) which should be used when
    printing this row/product.
    """)

    label_profile_uuid = sa.Column(sa.String(length=32), nullable=True)

    label_profile = orm.relationship(
        LabelProfile,
        doc="""
        Reference to the :class:`LabelProfile` with which the row is associated.
        """,
        backref=orm.backref('_print_labels_rows', cascade='all'))

    label_quantity = sa.Column(sa.Integer(), nullable=False, default=1, doc="""
    Number of labels to print for this row/product.
    """)

    ####################
    # product fields
    ####################

    location = sa.Column(sa.String(length=30), nullable=True)
    category_code = sa.Column(sa.String(length=30), nullable=True)
    category_name = sa.Column(sa.String(length=50), nullable=True)
    description_2 = sa.Column(sa.String(length=100), nullable=True)
    regular_price = sa.Column(sa.Numeric(precision=6, scale=2), nullable=True)
    pack_quantity = sa.Column(sa.Integer(), nullable=True)
    pack_price = sa.Column(sa.Numeric(precision=6, scale=2), nullable=True)

    sale_price = sa.Column(sa.Numeric(precision=6, scale=2), nullable=True)
    sale_start = sa.Column(sa.DateTime(), nullable=True)
    sale_stop = sa.Column(sa.DateTime(), nullable=True)

    tpr_price = sa.Column(sa.Numeric(precision=8, scale=3), nullable=True)
    tpr_starts = sa.Column(sa.DateTime(), nullable=True)
    tpr_ends = sa.Column(sa.DateTime(), nullable=True)

    current_price = sa.Column(sa.Numeric(precision=8, scale=3), nullable=True)
    current_starts = sa.Column(sa.DateTime(), nullable=True)
    current_ends = sa.Column(sa.DateTime(), nullable=True)

    multi_unit_sale_price = sa.Column(sa.String(length=20), nullable=True)
    comp_unit_size = sa.Column(sa.Numeric(precision=6, scale=2), nullable=True)
    comp_unit_measure = sa.Column(sa.String(length=30), nullable=True)
    comp_unit_reg_price = sa.Column(sa.String(length=10), nullable=True)
    comp_unit_sale_price = sa.Column(sa.String(length=10), nullable=True)
    vendor_id = sa.Column(sa.String(length=10), nullable=True)
    vendor_name = sa.Column(sa.String(length=50), nullable=True)
    vendor_item_code = sa.Column(sa.String(length=15), nullable=True)
    case_quantity = sa.Column(sa.Numeric(precision=6, scale=2), nullable=True)
    tax_code = sa.Column(sa.Integer(), nullable=True)
    crv = sa.Column(sa.Boolean(), nullable=True)
    organic = sa.Column(sa.Boolean(), nullable=True)
    country_of_origin = sa.Column(sa.String(length=100), nullable=True)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2019 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data model for generic "product" batches
"""

from __future__ import unicode_literals, absolute_import

import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.ext.declarative import declared_attr

from rattail.db.model import (Base, BatchMixin, ProductBatchRowMixin,
                              Vendor, Department, Subdepartment)
from rattail.db.core import filename_column


class ProductBatch(BatchMixin, Base):
    """
    Primary data model for product batches.
    """
    __tablename__ = 'batch_product'
    __batchrow_class__ = 'ProductBatchRow'
    batch_key = 'product'
    model_title = "Product Batch"
    model_title_plural = "Product Batches"

    input_filename = filename_column(nullable=True, doc="""
    Base name of the input data file, if applicable.
    """)


class ProductBatchRow(ProductBatchRowMixin, Base):
    """
    Row of data within a product batch.
    """
    __tablename__ = 'batch_product_row'
    __batch_class__ = ProductBatch

    @declared_attr
    def __table_args__(cls):
        return cls.__default_table_args__() + (
            sa.ForeignKeyConstraint(['department_uuid'], ['department.uuid'],
                                    name='batch_product_row_fk_department'),
            sa.ForeignKeyConstraint(['subdepartment_uuid'], ['subdepartment.uuid'],
                                    name='batch_product_row_fk_subdepartment'),
            sa.ForeignKeyConstraint(['category_uuid'], ['category.uuid'],
                                    name='batch_product_row_fk_category'),
            sa.ForeignKeyConstraint(['family_uuid'], ['family.uuid'],
                                    name='batch_product_row_fk_family'),
            sa.ForeignKeyConstraint(['reportcode_uuid'], ['report_code.uuid'],
                                    name='batch_product_row_fk_reportcode'),
            sa.ForeignKeyConstraint(['vendor_uuid'], ['vendor.uuid'],
                                    name='batch_product_row_fk_vendor'),
        )

    STATUS_OK                           = 1
    STATUS_MISSING_KEY                  = 2
    STATUS_PRODUCT_NOT_FOUND            = 3

    STATUS = {
        STATUS_OK                       : "ok",
        STATUS_MISSING_KEY              : "missing product key",
        STATUS_PRODUCT_NOT_FOUND        : "product not found",
    }

    department_uuid = sa.Column(sa.String(length=32), nullable=True)
    department = orm.relationship(Department, doc="""
    Reference to the department to which the product belongs.
    """)

    subdepartment_uuid = sa.Column(sa.String(length=32), nullable=True)
    subdepartment = orm.relationship(Subdepartment, doc="""
    Reference to the subdepartment to which the product belongs.
    """)

    vendor_uuid = sa.Column(sa.String(length=32), nullable=True)
    vendor = orm.relationship(Vendor, doc="""
    Reference to the "preferred" vendor from which product may be purchased.
    """)

    vendor_item_code = sa.Column(sa.String(length=20), nullable=True, doc="""
    Vendor-specific item code (SKU) for the product.
    """)

    category_uuid = sa.Column(sa.String(length=32), nullable=True)
    category = orm.relationship('Category', doc="""
    Reference to the category for the item.
    """)

    family_uuid = sa.Column(sa.String(length=32), nullable=True)
    family = orm.relationship('Family', doc="""
    Reference to the family for the item.
    """)

    reportcode_uuid = sa.Column(sa.String(length=32), nullable=True)
    reportcode = orm.relationship('ReportCode', doc="""
    Reference to the "report code" object for the item.
    """)

    regular_cost = sa.Column(sa.Numeric(precision=9, scale=5), nullable=True, doc="""
    The "base" unit cost for the item, i.e. with no discounts applied.
    """)

    current_cost = sa.Column(sa.Numeric(precision=9, scale=5), nullable=True, doc="""
    The "true" unit cost for the item, i.e. with discounts applied.
    """)

    current_cost_starts = sa.Column(sa.DateTime(), nullable=True, doc="""
    Date/time when the current cost starts, if applicable.
    """)

    current_cost_ends = sa.Column(sa.DateTime(), nullable=True, doc="""
    Date/time when the current cost ends, if applicable.
    """)

    regular_price = sa.Column(sa.Numeric(precision=8, scale=3), nullable=True, doc="""
    The "regular" unit price for the item.
    """)

    current_price = sa.Column(sa.Numeric(precision=8, scale=3), nullable=True, doc="""
    The "current" unit price for the item.
    """)

    current_price_starts = sa.Column(sa.DateTime(), nullable=True, doc="""
    Date/time when the current price starts, if applicable.
    """)

    current_price_ends = sa.Column(sa.DateTime(), nullable=True, doc="""
    Date/time when the current price ends, if applicable.
    """)

    suggested_price = sa.Column(sa.Numeric(precision=8, scale=3), nullable=True, doc="""
    The "suggested" retail price for the item.
    """)



# -*- coding: utf-8 -*-
"""add employee_x_store

Revision ID: 27bcaa3c7f5f
Revises: 57d12767eb53
Create Date: 2016-01-25 21:24:00.543909

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '27bcaa3c7f5f'
down_revision = u'57d12767eb53'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


def upgrade():

    # employee_x_store
    op.create_table('employee_x_store',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('employee_uuid', sa.String(length=32), nullable=False),
                    sa.Column('store_uuid', sa.String(length=32), nullable=False),
                    sa.ForeignKeyConstraint(['employee_uuid'], [u'employee.uuid'], name=u'employee_x_store_fk_employee'),
                    sa.ForeignKeyConstraint(['store_uuid'], [u'store.uuid'], name=u'employee_x_store_fk_store'),
                    sa.PrimaryKeyConstraint('uuid')
    )


def downgrade():

    # employee_x_store
    op.drop_table('employee_x_store')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Features
"""

from __future__ import unicode_literals, absolute_import

import json
from mako.lookup import TemplateLookup

from rattail.files import resource_path


class Feature(object):
    """
    Base class for features.
    """

    def __init__(self, config):
        self.config = config

        # TODO: make templates dir configurable
        templates = [resource_path('rattail:templates/feature')]
        self.templates = TemplateLookup(directories=templates)

    def get_defaults(self):
        return {}

    def get_template(self, path):
        """
        Locate and return the given Mako feature template.
        """
        return self.templates.get_template(path)

    def generate_mako(self, template, **context):
        """
        Generate and return output from the given template.
        """
        template = self.get_template(template)
        text = template.render(**context)
        return text


class JSONWidget(object):
    """
    This is a minimal "widget" for use with certain colander schema nodes,
    which lets us accept a JSON-encoded string, whereas otherwise Deform would
    be expecting a peppercorn-compatible series of field mappings.
    """

    def deserialize(self, field, pstruct):
        return json.loads(pstruct)



# -*- coding: utf-8; -*-
"""add receiving_complete flag

Revision ID: c45017b36f25
Revises: fd0975329091
Create Date: 2019-11-26 13:39:30.080843

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = 'c45017b36f25'
down_revision = 'fd0975329091'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # purchase_batch
    op.add_column('purchase_batch', sa.Column('receiving_complete', sa.Boolean(), nullable=True))


def downgrade():

    # purchase_batch
    op.drop_column('purchase_batch', 'receiving_complete')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2022 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Problem Report Handlers
"""

from __future__ import unicode_literals, absolute_import

import logging

from rattail.db import Session
from rattail.mail import send_email
from rattail.time import localtime
from rattail.util import load_object, import_module_path, progress_loop
from rattail.problems import ProblemReport, RattailProblemReport


log = logging.getLogger(__name__)


class ProblemReportHandler(object):
    """
    Base class and default implementation for problem report handlers.
    """

    def __init__(self, config, dry_run=False, progress=None):
        self.config = config
        self.dry_run = dry_run
        self.progress = progress
        self.app = self.config.get_app()
        self.enum = self.config.get_enum()

    def progress_loop(self, func, items, factory=None, **kwargs):
        factory = factory or self.progress
        return progress_loop(func, items, factory, **kwargs)

    def get_all_problem_reports(self):
        """
        Returns a simple list of all ``ProblemReport`` subclasses which are
        "available" according to config.
        """
        reports = []
        problem_modules = self.config.getlist('rattail', 'problems',
                                              default=['rattail.problems.rattail'])
        for module_path in problem_modules:
            module = import_module_path(module_path)
            for name in dir(module):
                obj = getattr(module, name)

                if (isinstance(obj, type) and
                    issubclass(obj, ProblemReport) and
                    obj not in (ProblemReport, RattailProblemReport)):

                    reports.append(obj)

        return reports

    def get_problem_reports(self, systems=None, problems=None):
        """
        Return a list of all problem reports which match the given criteria.

        :param systems: Optional list of "system keys" which a problem report
           must match, in order to be included in return value.

        :param problems: Optional list of "problem keys" which a problem report
           must match, in order to be included in return value.

        :returns: List of problem reports; may be an empty list.
        """
        all_reports = self.get_all_problem_reports()
        if not (systems or problems):
            return all_reports

        matches = []
        for report in all_reports:
            if not systems or report.system_key in systems:
                if not problems or report.problem_key in problems:
                    matches.append(report)
        return matches

    def get_problem_report(self, system_key, problem_key, **kwargs):
        """
        Return a specific problem report, identified by the
        system/problem key pair.

        :param system_key: System key, part of the identifier.

        :param problem_key: Problem key, part of the identifier.

        :returns: A specific :class:`~ProblemReport` class, if a match
           was found, otherwise ``None``.
        """
        reports = self.get_problem_reports(systems=[system_key],
                                           problems=[problem_key])
        if reports:
            if len(reports) > 1:
                raise RuntimeError("Multiple problem reports defined "
                                   "for key: {}.{}".format(
                                       system_key, problem_key))
            return reports[0]

    def organize_problem_reports(self, reports):
        """
        Returns a nested dict with the given problem reports.
        """
        organized = {}

        for report in reports:
            system = organized.setdefault(report.system_key, {})
            system[report.problem_key] = report

        return organized

    def supported_systems(self):
        """
        Returns list of keys for all systems which are supported by any of the
        available problem reports, according to config.
        """
        problem_reports = self.get_all_problem_reports()
        return sorted(set([pr.system_key for pr in problem_reports]))

    def run_problem_reports(self, reports, fix=False):
        """
        Run the given set of problem reports.

        :param fix: This flag will be passed as-is to
           :meth:`run_problem_report()`.
        """
        organized = self.organize_problem_reports(reports)
        for system_key in sorted(organized):
            system = organized[system_key]
            for problem_key in sorted(system):
                report = system[problem_key]
                self.run_problem_report(report, fix=fix)

    def run_problem_report(self, problem_report, fix=False,
                           send=True, **kwargs):
        """
        Run a specific problem report.
        """
        log.info("running problem report: %s.%s",
                 problem_report.system_key,
                 problem_report.problem_key)

        progress = kwargs.pop('progress', self.progress)
        report = problem_report(self.config,
                                dry_run=self.dry_run,
                                progress=progress)

        problems = report.find_problems(**kwargs)
        log.info("found %s problems", len(problems))
        if problems and send:
            self.send_problem_report(report, problems)
        return problems

    def get_email_key(self, report):
        if report.email_key:
            return report.email_key

        return '{}_problems_{}'.format(report.system_key,
                                       report.problem_key)

    def send_problem_report(self, report, problems):
        """
        Send out an email with details of the given problem report.
        """
        context = self.get_global_email_context()
        context = self.get_report_email_context(report, problems, **context)
        context.update({
            'report': report,
            'problems': problems,
            'app': self.app,
            'enum': self.enum,
            'render_time': self.render_time,
        })

        email_key = self.get_email_key(report)
        send_email(self.config, email_key, context,
                   default_subject=report.problem_title)

    def render_time(self, time, from_utc=True):
        """
        Render the given timestamp, localizing if necessary.
        """
        time = self.app.localtime(time, from_utc=from_utc)
        return self.app.render_datetime(time)

    def get_global_email_context(self, **kwargs):
        """
        This method can be used to add extra context for all email templates.
        """
        return kwargs

    def get_report_email_context(self, report, problems, **kwargs):
        """
        This method can be used to add extra context for a specific report's
        email template.
        """
        kwargs['system_title'] = self.get_system_title(report.system_key)
        kwargs = report.get_email_context(problems, **kwargs)
        return kwargs

    def get_system_title(self, system_key):
        """
        Should return a "display title" for the given system.
        """
        return system_key.capitalize()


def get_problem_report_handler(config, **kwargs):
    """
    Create and return the configured :class:`ProblemReportHandler` instance.
    """
    spec = config.get('rattail.problems', 'handler')
    if spec:
        factory = load_object(spec)
    else:
        factory = ProblemReportHandler
    return factory(config, **kwargs)



# -*- coding: utf-8 -*-
"""add purchase_batch.mode

Revision ID: 4a4ae7ebdae6
Revises: 19da5ebe6fb5
Create Date: 2016-11-20 16:56:24.788146

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '4a4ae7ebdae6'
down_revision = u'19da5ebe6fb5'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql
from rattail import enum


batch_id_seq = sa.Sequence('batch_id_seq')


def upgrade():

    # purchase_batch
    op.add_column('purchase_batch', sa.Column('purchase_uuid', sa.String(length=32), nullable=True))
    op.create_foreign_key(u'purchase_batch_fk_purchase', 'purchase_batch', 'purchase', ['purchase_uuid'], ['uuid'])

    # purchase_batch.mode (provide default value then disallow null)
    op.add_column('purchase_batch', sa.Column('mode', sa.Integer(), nullable=True))
    batch = sa.sql.table('purchase_batch', sa.sql.column('mode'))
    op.execute(batch.update().values({'mode': enum.PURCHASE_BATCH_MODE_ORDERING}))
    op.alter_column('purchase_batch', 'mode', nullable=False,
                    existing_type=sa.Integer())


def downgrade():

    # purchase_batch
    op.drop_constraint(u'purchase_batch_fk_purchase', 'purchase_batch', type_='foreignkey')
    op.drop_column('purchase_batch', 'purchase_uuid')
    op.drop_column('purchase_batch', 'mode')



# -*- coding: utf-8; -*-
"""add pricing_batch.filename

Revision ID: c02430f167c2
Revises: f7dc4e92752c
Create Date: 2018-12-17 22:05:06.558470

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = 'c02430f167c2'
down_revision = 'f7dc4e92752c'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # batch_pricing
    op.add_column('batch_pricing', sa.Column('input_filename', sa.String(length=255), nullable=True))


def downgrade():

    # batch_pricing
    op.drop_column('batch_pricing', 'input_filename')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2018 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data Models for Email Bouncer Daemon
"""

from __future__ import unicode_literals, absolute_import

import datetime

import six
import sqlalchemy as sa
from sqlalchemy import orm

from rattail.db.model import Base, uuid_column
from rattail.db.model import User


class EmailAttempt(Base):
    """
    Represents an attempt to send an email.  Ideally the generated message body
    is persisted on disk, in addition to these details.
    """
    __tablename__ = 'email_attempt'

    uuid = uuid_column()

    key = sa.Column(sa.String(length=254), nullable=False, doc="""
    Config key for the email being sent.
    """)

    sender = sa.Column(sa.Text(), nullable=False, doc="""
    Value for the From: email header.
    """)

    to = sa.Column(sa.Text(), nullable=False, doc="""
    Value for the To: email header.
    """)

    cc = sa.Column(sa.Text(), nullable=True, doc="""
    Value for the Cc: email header.
    """)

    bcc = sa.Column(sa.Text(), nullable=True, doc="""
    Value for the Bcc: email header.
    """)

    subject = sa.Column(sa.String(length=254), nullable=False, default='', doc="""
    Value for the Subject: email header.
    """)

    sent = sa.Column(sa.DateTime(), nullable=False, doc="""
    Date and time when email send attempt was first made.
    """)

    status_code = sa.Column(sa.Integer(), nullable=False, doc="""
    Status code for the email.
    """)

    status_text = sa.Column(sa.Text(), nullable=True, doc="""
    Status text for the email, if applicable.
    """)

    def __str__(self):
        if six.PY2:
            return self.subject.encode('utf8')
        return self.subject

    if six.PY2:
        def __unicode__(self):
            return self.subject


@six.python_2_unicode_compatible
class EmailBounce(Base):
    """
    Represents an email bounce notification message.  This table is populated
    by the Bouncer daemon and then exposed as a workflow queue within Tailbone.
    """
    __tablename__ = 'email_bounce'
    __table_args__ = (
        sa.ForeignKeyConstraint(['processed_by_uuid'], ['user.uuid'], name='email_bounce_fk_processed_by'),
        )

    uuid = uuid_column()

    config_key = sa.Column(sa.String(length=20), nullable=False, doc="""
    Key for the configuration profile with which the bounce is associated.
    This profile is also what determines the handler for the bounce.
    """)

    bounced = sa.Column(sa.DateTime(), nullable=False, default=datetime.datetime.utcnow, doc="""
    Date and time when the email bounce was discovered.  Defaults to current time.
    """)

    bounce_recipient_address = sa.Column(sa.String(length=255), nullable=False, doc="""
    Email address to which the bounce notification message was sent.
    """)

    intended_recipient_address = sa.Column(sa.String(length=255), nullable=True, doc="""
    Email address of the original intended recipient, if one could be determined.
    """)

    intended_recipient_key = sa.Column(sa.String(length=20), nullable=True, doc="""
    Generic key for the intended recipient.  This must be populated and
    interpreted by a custom bounce handler.
    """)

    processed = sa.Column(sa.DateTime(), nullable=True, doc="""
    Date and time when the email bounce was fully processed by a user.
    """)

    processed_by_uuid = sa.Column(sa.String(length=32), nullable=True)

    processed_by = orm.relationship(User, doc="""
    Reference to the :class:`rattail.db.model.User` who processed the bounce.
    """)

    def __str__(self):
        if self.intended_recipient_address:
            return str(self.intended_recipient_address)
        return str(self.bounce_recipient_address)



# -*- coding: utf-8; -*-
"""add role.sync_users

Revision ID: 678a32b6cb19
Revises: 43b9e0a6c14e
Create Date: 2021-11-13 14:52:37.243794

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '678a32b6cb19'
down_revision = '43b9e0a6c14e'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # role
    op.add_column('role', sa.Column('sync_users', sa.Boolean(), nullable=True))
    op.add_column('role_version', sa.Column('sync_users', sa.Boolean(), autoincrement=False, nullable=True))


def downgrade():

    # role
    op.drop_column('role_version', 'sync_users')
    op.drop_column('role', 'sync_users')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
MailMon Utilities
"""

from __future__ import unicode_literals, absolute_import

import datetime

from rattail.db import api
from rattail.time import localtime, make_utc


def get_lastrun_setting(config, watcher_key):
    """
    Retrieve the "last run" setting name for the given watcher key.
    """
    return 'rattail.mailmon.{}.watcher.lastrun'.format(watcher_key)


def get_lastrun_timefmt(config):
    """
    Retrieve the "last run" time format.
    """
    return '%Y-%m-%d %H:%M:%S'


def get_lastrun(config, watcher_key, local=False, tzinfo=True, session=None):
    """
    Retrieve the "last run" time for the mailmon watcher thread
    identifed by ``watcher_key``.
    """
    # the 'last run' value is maintained as UTC
    lastrun_setting = get_lastrun_setting(config, watcher_key)
    timefmt = get_lastrun_timefmt(config)
    lastrun = api.get_setting(session, lastrun_setting)
    if lastrun:
        lastrun = datetime.datetime.strptime(lastrun, timefmt)
        if local:
            return localtime(config, lastrun, from_utc=True, tzinfo=tzinfo)
        else:
            return make_utc(lastrun, tzinfo=tzinfo)



# -*- coding: utf-8 -*-
"""add customer number

Revision ID: 195ceb6abeb3
Revises: 6a1ec8b93637
Create Date: 2017-05-11 11:29:13.525277

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '195ceb6abeb3'
down_revision = u'6a1ec8b93637'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


def upgrade():

    # customer
    op.add_column('customer', sa.Column('number', sa.Integer(), nullable=True))


def downgrade():

    # customer
    op.drop_column('customer', 'number')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2018 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Global Product Code
"""

from __future__ import unicode_literals, absolute_import

import six

from rattail import barcodes


@six.python_2_unicode_compatible
class GPC(object):
    """
    Class to abstract the details of Global Product Code data.  Examples of
    this would be UPC or EAN barcodes.

    The initial motivation for this class was to provide better SIL support.
    To that end, the instances are assumed to always be comprised of only
    numeric digits, and must include a check digit.  If you do not know the
    check digit, provide a ``calc_check_digit`` value to the constructor.
    """

    def __init__(self, value, calc_check_digit=False):
        """
        Constructor.  ``value`` must be either an integer or a long value, or a
        string containing only digits.

        If ``calc_check_digit`` is ``False``, then ``value`` is assumed to
        include the check digit.  If the value does not include a check digit
        and needs one to be calculated, then ``calc_check_digit`` should be a
        keyword signifying the algorithm to be used.

        Currently the only check digit algorithm keyword supported is
        ``'upc'``.  As that is likely to always be the default, a
        ``calc_check_digit`` value of ``True`` will be perceived as equivalent
        to ``'upc'``.
        """
        value = six.text_type(value)

        if calc_check_digit == 'auto':
            calc_check_digit = 'upc' if len(value) < 12 else False

        if calc_check_digit is True or calc_check_digit == 'upc':
            value += six.text_type(barcodes.upc_check_digit(value))

        self.value = int(value)

    def __eq__(self, other):
        try:
            return int(self) == int(other)
        except (TypeError, ValueError):
            return False

    def __ne__(self, other):
        try:
            return int(self) != int(other)
        except (TypeError, ValueError):
            return True

    # TODO: this is no longer used in python3
    # https://docs.python.org/3/whatsnew/3.0.html#ordering-comparisons
    # TODO: need to implement "rich comparisons" instead
    # https://docs.python.org/3/reference/datamodel.html#object.__lt__
    def __cmp__(self, other):

        # treat non-integers as being less than myself
        try:
            other = int(other)
        except (TypeError, ValueError):
            return 1

        myself = int(self)
        if myself < other:
            return -1
        if myself > other:
            return 1
        assert myself == other
        return 0

    def __lt__(self, other):
        try:
            return int(self) < int(other)
        except (TypeError, ValueError):
            return False

    def __hash__(self):
        return hash(self.value)

    def __int__(self):
        return int(self.value)

    def __long__(self):
        return long(self.value)

    def __repr__(self):
        return "GPC('%014d')" % self.value

    def __str__(self):
        # return str('%14d' % self.value)
        return str('{:014d}'.format(self.value))

    @property
    def data_str(self):
        """
        Returns the "data" for the barcode as unicode string, i.e. minus check
        digit and also with all leading zeroes removed.  A visual is maybe
        helpful here::

           >>> upc = GPC('7430500132', calc_check_digit='upc')
           >>> print(repr(upc))
           GPC('00074305001321')
           >>> print(repr(upc.data_str))
           u'7430500132'

        Note that in this case the ``data_str`` value is the same as was
        originally provided to the constructor, but that isn't always the case.
        """
        return six.text_type(int(self))[:-1]

    @property
    def data_length(self):
        """
        Returns the length of the "data" for the barcode.  This is just a
        convenience that returns ``len(self.data_str)``.
        """
        return len(self.data_str)

    @property
    def type2_upc(self):
        """
        Returns boolean indicating whether the barcode has "type 2" UPC data.
        """
        return self.data_str.startswith('2') and self.data_length == 11

    def pretty(self):
        """
        Returns the UPC as a somewhat more human-readable string.  Basically
        that just means the check digit is distinguished by a hyphen.
        """
        upc = six.text_type(self)
        return "{0}-{1}".format(upc[:-1], upc[-1])



# -*- coding: utf-8; -*-
"""add vendor id, name for catalog batch

Revision ID: 90facf64f153
Revises: 387ba7573623
Create Date: 2021-02-14 12:04:17.321423

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '90facf64f153'
down_revision = '387ba7573623'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # batch_vendorcatalog
    op.add_column('batch_vendorcatalog', sa.Column('vendor_id', sa.String(length=15), nullable=True))
    op.add_column('batch_vendorcatalog', sa.Column('vendor_name', sa.String(length=50), nullable=True))
    op.alter_column('batch_vendorcatalog', 'vendor_uuid',
               existing_type=sa.VARCHAR(length=32),
               nullable=True)


def downgrade():

    # batch_vendorcatalog
    # TODO: cannot necessarily reverse this one..?
    op.alter_column('batch_vendorcatalog', 'vendor_uuid',
               existing_type=sa.VARCHAR(length=32),
               nullable=False)
    op.drop_column('batch_vendorcatalog', 'vendor_name')
    op.drop_column('batch_vendorcatalog', 'vendor_id')



# -*- coding: utf-8 -*-
"""add upgrades

Revision ID: e5c3eab67b28
Revises: 1242c7a0e24a
Create Date: 2017-08-05 16:30:36.615368

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = 'e5c3eab67b28'
down_revision = u'1242c7a0e24a'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # upgrade
    op.create_table('upgrade',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('created', sa.DateTime(), nullable=True),
                    sa.Column('created_by_uuid', sa.String(length=32), nullable=False),
                    sa.Column('description', sa.String(length=255), nullable=False),
                    sa.Column('not_until', sa.DateTime(), nullable=True),
                    sa.Column('notes', sa.Text(), nullable=True),
                    sa.Column('enabled', sa.Boolean(), nullable=False),
                    sa.Column('executing', sa.Boolean(), nullable=False),
                    sa.Column('executed', sa.DateTime(), nullable=True),
                    sa.Column('executed_by_uuid', sa.String(length=32), nullable=True),
                    sa.ForeignKeyConstraint(['created_by_uuid'], [u'user.uuid'], name=u'upgrade_fk_created_by'),
                    sa.ForeignKeyConstraint(['executed_by_uuid'], [u'user.uuid'], name=u'upgrade_fk_executed_by'),
                    sa.PrimaryKeyConstraint('uuid')
    )

    # upgrade_requirement
    op.create_table('upgrade_requirement',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('upgrade_uuid', sa.String(length=32), nullable=False),
                    sa.Column('package', sa.String(length=255), nullable=False),
                    sa.Column('version', sa.String(length=255), nullable=False),
                    sa.ForeignKeyConstraint(['upgrade_uuid'], [u'upgrade.uuid'], name=u'upgrade_requirement_fk_upgrade'),
                    sa.PrimaryKeyConstraint('uuid')
    )


def downgrade():

    # upgrade*
    op.drop_table('upgrade_requirement')
    op.drop_table('upgrade')



# -*- coding: utf-8; -*-
"""add datasync_change.batch_id

Revision ID: fd91ee24cead
Revises: 7d0c93d3ff1e
Create Date: 2020-01-15 17:34:22.597552

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = 'fd91ee24cead'
down_revision = '7d0c93d3ff1e'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types


batch_id_seq = sa.Sequence('datasync_change_batch_id_seq')


def upgrade():
    from sqlalchemy.schema import CreateSequence

    # datasync_change
    op.add_column('datasync_change', sa.Column('batch_id', sa.Integer(), nullable=True))
    op.add_column('datasync_change', sa.Column('batch_sequence', sa.Integer(), nullable=True))
    op.execute(CreateSequence(batch_id_seq))


def downgrade():
    from sqlalchemy.schema import DropSequence

    # datasync_change
    op.drop_column('datasync_change', 'batch_sequence')
    op.drop_column('datasync_change', 'batch_id')
    op.execute(DropSequence(batch_id_seq))



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2022 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Batch-related commands
"""

from __future__ import unicode_literals, absolute_import

import os
import datetime
import logging

from rattail.commands.core import Subcommand, date_argument, dict_argument
from rattail.progress import SocketProgress
from rattail.time import localtime, make_utc


log = logging.getLogger(__name__)


class BatchHandlerCommand(Subcommand):
    """
    Base class for commands which invoke a batch handler.
    """

    def add_parser_args(self, parser):
        parser.add_argument('--batch-type', metavar='KEY',
                            help="Type of batch to be dealt with, e.g. 'vendor_catalog'")
        parser.add_argument('--dry-run', action='store_true',
                            help="Go through the full motions and allow logging etc. to "
                            "occur, but rollback (abort) the transaction at the end.")

    def get_handler(self, args):
        """
        Must return the batch handler to use, per the given ``args``
        object.

        Default logic assumes that ``args`` has a ``batch_type``
        attribute, which determines which batch handler should be
        used.  That is figured out by calling
        :meth:`~rattail.app.AppHandler.get_batch_handler()` on the app
        handler.

        :param args: Reference to the
           :class:`python:argparse.Namespace` instance, which was the
           result of parsing the command line args.

        :returns: Must return the batch handler.  This will be an
           instance of some class which derives from
           :class:`~rattail.batch.handlers.BatchHandler`.
        """
        return self.app.get_batch_handler(args.batch_type)

    def run(self, args):
        handler = self.get_handler(args)
        session = self.make_session()
        user = self.get_runas_user(session)

        success = self.do_stuff(args, handler, session, user)

        if args.dry_run:
            session.rollback()
            log.info("dry run, so transaction was rolled back")
        elif success:
            session.commit()
            log.info("transaction was committed")
        else:
            session.rollback()
            log.warning("action failed, so transaction was rolled back")
        session.close()

    def do_stuff(self, args, handler, session, user):
        raise NotImplementedError


class MakeBatch(BatchHandlerCommand):
    """
    Make a new batch, from a data file
    """
    name = 'make-batch'
    description = __doc__.strip()

    def add_parser_args(self, parser):
        super(MakeBatch, self).add_parser_args(parser)

        parser.add_argument('--input-file', metavar='PATH', required=True,
                            help="Path to single input file, to be used as data "
                            "source for the new batch.  (File format will vary "
                            "depending on batch type.)")

        parser.add_argument('--kwargs', type=dict_argument, default={},
                            help="Optional JSON-encoded string containing extra "
                            "keyword arguments to be passed to the handler's batch "
                            "creation logic.")

    def do_stuff(self, args, handler, session, user):
        """
        This will create a new batch of the specified type, then populate it
        with the given data file.
        """
        if not os.path.exists(args.input_file):
            raise RuntimeError("input file path does not exist: {}".format(args.input_file))

        kwargs = dict(args.kwargs)
        delete_if_empty = kwargs.pop('delete_if_empty', False)
        auto_execute_allowed = kwargs.pop('auto_execute_allowed', True)

        batch = handler.make_batch(session, created_by=user, **kwargs)
        handler.set_input_file(batch, args.input_file)
        handler.do_populate(batch, user)

        if delete_if_empty:
            session.flush()
            if not batch.data_rows:
                log.debug("auto-deleting empty '%s' batch: %s", handler.batch_key, batch)
                handler.do_delete(batch, dry_run=args.dry_run,
                                  progress=self.progress)
                batch = None

        if batch and auto_execute_allowed and handler.auto_executable(batch):
            handler.execute(batch, user=user)
            batch.executed = make_utc()
            batch.executed_by = user

        return True


class BatchAction(BatchHandlerCommand):
    """
    Base class for commands which invoke a handler to act on a single batch.
    """

    def add_parser_args(self, parser):
        super(BatchAction, self).add_parser_args(parser)
        parser.add_argument('batch_uuid',
                            help="UUID of the batch to be populated.")

    def do_stuff(self, args, handler, session, user):
        """
        This will invoke some action on the batch.
        """
        batch = session.query(handler.batch_model_class).get(args.batch_uuid)
        if not batch:
            raise RuntimeError("Batch of type '{}' not found: {}".format(args.batch_type, args.batch_uuid))

        return self.action(args, handler, batch, user)


class PopulateBatch(BatchAction):
    """
    Populate initial data for a batch
    """
    name = 'populate-batch'
    description = __doc__.strip()

    def action(self, args, handler, batch, user):
        return handler.do_populate(batch, user, progress=self.progress)


class RefreshBatch(BatchAction):
    """
    Refresh data for a batch
    """
    name = 'refresh-batch'
    description = __doc__.strip()

    def action(self, args, handler, batch, user):
        return handler.do_refresh(batch, user, progress=self.progress)


class ExecuteBatch(BatchAction):
    """
    Execute a batch
    """
    name = 'execute-batch'
    description = __doc__.strip()

    def add_parser_args(self, parser):
        super(ExecuteBatch, self).add_parser_args(parser)

        parser.add_argument('--kwargs', type=dict_argument, default={},
                            help="Optional JSON-encoded string containing extra "
                            "keyword arguments to be passed to the handler's batch "
                            "execution function.")

    def action(self, args, handler, batch, user):
        return handler.do_execute(batch, user, progress=self.progress, **args.kwargs)


class PurgeBatches(BatchHandlerCommand):
    """
    Purge old batches from the database
    """
    name = 'purge-batches'
    description = __doc__.strip()

    def add_parser_args(self, parser):
        super(PurgeBatches, self).add_parser_args(parser)

        parser.add_argument('--before', type=date_argument, metavar='DATE',
                            help="Purge all batches executed prior to this date.  If not "
                            "specified, will use --before-days to calculate instead.")

        parser.add_argument('--before-days', type=int, default=90, metavar='DAYS',
                            help="Number of days before the current date, to be used "
                            "as the cutoff date if --before is not specified.  Default "
                            "is 90 days before current date.")

        parser.add_argument('--list-types', action='store_true',
                            help="If set, list available batch types instead of trying "
                            "to purge anything.")


    def run(self, args):
        if args.list_types:
            self.list_types()
            return

        handler = self.get_handler(args)
        session = self.make_session()

        kwargs = {
            'dry_run': args.dry_run,
            'progress': self.progress,
        }
        if args.before:
            before = datetime.datetime.combine(args.before, datetime.time(0))
            before = localtime(self.config, before)
            kwargs['before'] = before
        else:
            kwargs['before_days'] = args.before_days

        purged = handler.purge_batches(session, **kwargs)

        if args.dry_run:
            session.rollback()
            log.info("dry run, so transaction was rolled back")
        else:
            session.commit()
            log.info("transaction was committed")
        session.close()

    def list_types(self):
        from rattail.batch.handlers import get_batch_types

        keys = get_batch_types(self.config)
        for key in keys:
            self.stdout.write("{}\n".format(key))



# -*- coding: utf-8 -*-
"""add employee_history

Revision ID: 7f88949a7ab9
Revises: eb4e965ce771
Create Date: 2017-10-17 08:26:31.693081

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '7f88949a7ab9'
down_revision = u'eb4e965ce771'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # employee_history
    op.create_table('employee_history',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('employee_uuid', sa.String(length=32), nullable=False),
                    sa.Column('start_date', sa.Date(), nullable=False),
                    sa.Column('end_date', sa.Date(), nullable=True),
                    sa.ForeignKeyConstraint(['employee_uuid'], [u'employee.uuid'], name=u'employee_history_fk_employee'),
                    sa.PrimaryKeyConstraint('uuid')
    )
    op.create_table('employee_history_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('employee_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('start_date', sa.Date(), autoincrement=False, nullable=True),
                    sa.Column('end_date', sa.Date(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_employee_history_version_end_transaction_id'), 'employee_history_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_employee_history_version_operation_type'), 'employee_history_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_employee_history_version_transaction_id'), 'employee_history_version', ['transaction_id'], unique=False)


def downgrade():

    # employee_history
    op.drop_index(op.f('ix_employee_history_version_transaction_id'), table_name='employee_history_version')
    op.drop_index(op.f('ix_employee_history_version_operation_type'), table_name='employee_history_version')
    op.drop_index(op.f('ix_employee_history_version_end_transaction_id'), table_name='employee_history_version')
    op.drop_table('employee_history_version')
    op.drop_table('employee_history')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2022 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Handheld batch handler
"""

from __future__ import unicode_literals, absolute_import

import csv
import decimal

import six

from rattail.db import api, model
from rattail.batch import BatchHandler
from rattail.gpc import GPC
from rattail.time import make_utc
from rattail.wince import parse_batch_file as parse_wince_file


class HandheldBatchHandler(BatchHandler):
    """
    Handler for handheld batches.
    """
    batch_model_class = model.HandheldBatch

    def should_populate(self, batch):
        # all handheld batches must come from input data file
        return True

    def populate(self, batch, progress=None):
        """
        Pre-fill batch with row data from an input data file, parsed according
        to the batch device type.
        """
        if not batch.filename:
            raise ValueError("Batch must have a filename: {}".format(batch))
        if not batch.device_type:
            raise ValueError("Batch must have a device_type: {}".format(batch))
        batch.rowcount = 0

        def append(entry, i):
            upc, cases, units = entry
            row = model.HandheldBatchRow(upc=upc, cases=cases, units=units)
            batch.add_row(row)
            self.refresh_row(row)
            batch.rowcount += 1

        parse = getattr(self, 'parse_input_file_{}'.format(batch.device_type))
        entries = parse(batch.absolute_filepath(self.config), progress=progress)
        self.progress_loop(append, entries, progress,
                           message="Adding initial rows to batch")

    def refresh_batch_status(self, batch):
        if any([row.status_code != row.STATUS_OK for row in batch.active_rows()]):
            batch.status_code = batch.STATUS_QUESTIONABLE
        else:
            batch.status_code = batch.STATUS_OK

    def parse_input_file_motorola(self, path, progress=None):
        """
        Parse a RattailCE (binary or CSV) file to generate initial rows.
        """
        data = []
        filemode = 'rt' if six.PY3 else 'rb'
        with open(path, filemode) as f:
            line = f.readline()

        if '\x00' in line:      # raw binary file from RattailCE app

            def convert(entry, i):
                scancode, cases, units = entry
                upc = GPC(int(scancode), calc_check_digit='upc')
                data.append((
                    upc,
                    cases or None,
                    units or None,
                ))

            entries = list(parse_wince_file(path, progress=progress))

        else:                   # presumably csv, converted from raw file

            def convert(entry, i):
                upc = GPC(entry['upc'], calc_check_digit='upc')
                data.append((
                    upc,
                    decimal.Decimal(entry['cases']) if entry['cases'] else None,
                    decimal.Decimal(entry['units']) if entry['units'] else None,
                ))

            with open(path, filemode) as f:
                reader = csv.DictReader(f)
                entries = list(reader)

        self.progress_loop(convert, entries, progress,
                           message="Normalizing data from WinCE file")
        return data

    def parse_input_file_palmos(self, path, progress=None):
        """
        Parse a Rattail PalmOS (CSV) file to generate initial rows.
        """
        data = []

        def convert(entry, i):
            data.append((
                GPC(entry['upc'], calc_check_digit='upc'),
                int(entry['cases']),
                int(entry['units']),
            ))

        with open(path, 'rb') as f:
            reader = csv.DictReader(f)
            entries = list(reader)

        if self.progress_loop(convert, entries, progress,
                              message="Normalizing data from PalmOS file"):
            return data

    def refresh_row(self, row):
        """
        This method will be passed a row object which has already been properly
        added to a batch, and which has basic required fields already
        populated.  This method is then responsible for further populating all
        applicable fields for the row, based on current data within the
        relevant system(s).

        Note that in some cases this method may be called multiple times for
        the same row, e.g. once when first creating the batch and then later
        when a user explicitly refreshes the batch.  The method logic must
        account for this possibility.
        """
        if not row.upc:
            row.status_code = row.STATUS_PRODUCT_NOT_FOUND
            return

        session = self.app.get_session(row)
        product = api.get_product_by_upc(session, row.upc)
        if not product:
            row.status_code = row.STATUS_PRODUCT_NOT_FOUND
            return

        # current / static attributes
        row.product = product
        row.brand_name = product.brand.name if product.brand else None
        row.description = product.description
        row.size = product.size
        row.status_code = row.STATUS_OK

    def describe_execution(self, batch, **kwargs):
        return ("A new batch will be created, using the items from this one.  "
                "Type of the new batch depends on your choice of action.")

    def execute(self, batch, user=None, action='make_inventory_batch', progress=None, **kwargs):
        return self.execute_many([batch], user=user, action=action, progress=progress, **kwargs)

    def execute_many(self, batches, user=None, action='make_inventory_batch', progress=None, **kwargs):
        batches = [batch for batch in batches if not batch.executed]
        if not batches:
            return True
        if action == 'make_inventory_batch':
            result = self.make_inventory_batch(batches, user, progress=progress)
        elif action == 'make_label_batch':
            result = self.make_label_batch(batches, user, progress=progress)
        else:
            raise RuntimeError("Batch execution action is not supported: {}".format(action))
        now = make_utc()
        for batch in batches:
            batch.executed = now
            batch.executed_by = user
        return result

    def make_inventory_batch(self, handheld_batches, user, progress=None,
                             **kwargs):
        """
        Make a new Inventory Batch from the given Handheld Batch(es).

        :param handheld_batches: Sequence of one or more
           :class:`~rattail.db.model.batch.handheld.HandheldBatch`
           instances from which a new inventory batch should be made.

        :param user: :class:`~rattail.db.model.users.User` who is
           responsible for this action.

        :returns: A new
           :class:`~rattail.db.model.batch.inventory.InventoryBatch`
           instance, populated from the given handheld batch(es).
        """
        handler = self.app.get_batch_handler(
            'inventory', default='rattail.batch.inventory:InventoryBatchHandler')
        session = self.app.get_session(handheld_batches[0])
        batch = handler.make_batch(session, created_by=user,
                                   handheld_batches=handheld_batches,
                                   **kwargs)
        handler.do_populate(batch, user, progress=progress)
        return batch

    def make_label_batch(self, handheld_batches, user, progress=None,
                         **kwargs):
        """
        Make a new Label Batch from the given Handheld Batch(es).

        :param handheld_batches: Sequence of one or more
           :class:`~rattail.db.model.batch.handheld.HandheldBatch`
           instances from which a new label batch should be made.

        :param user: :class:`~rattail.db.model.users.User` who is
           responsible for this action.

        :returns: A new
           :class:`~rattail.db.model.batch.labels.LabelBatch`
           instance, populated from the given handheld batch(es).
        """
        handler = self.app.get_batch_handler(
            'labels', default='rattail.batch.labels:LabelBatchHandler')
        session = self.app.get_session(handheld_batches[0])
        if len(handheld_batches) > 1:
            # TODO: need to implement this
            raise NotImplementedError("Multiple handheld batches not (yet) "
                                      "supported when converting to label batch.  "
                                      "Please try again with only a single "
                                      "handheld batch")
        batch = handler.make_batch(session, created_by=user,
                                   handheld_batch=handheld_batches[0],
                                   **kwargs)
        handler.do_populate(batch, user, progress=progress)
        return batch



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2022 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Handler for Vendor Catalog batches
"""

from __future__ import unicode_literals, absolute_import

import decimal

from sqlalchemy import orm

from rattail.db import model
from rattail.batch import BatchHandler


class VendorCatalogHandler(BatchHandler):
    """
    Handler for vendor catalog batches.
    """
    batch_model_class = model.VendorCatalogBatch

    # make sure web app knows to employ versioning workarounds
    # TODO: i am actually not sure why these *always* seem to be needed for
    # this batch?  maybe the product relationships are too "tight" somehow?
    populate_with_versioning = False
    refresh_with_versioning = False
    execute_with_versioning = False

    version_catchup_execute = [
        'ProductCost',
    ]

    # can set these to e.g. Decimal('0.01') to "ignore" cost diffs below that
    case_cost_diff_threshold = None
    unit_cost_diff_threshold = None

    def allow_future(self):
        """
        Returns boolean indicating whether "future" cost changes
        should be allowed.

        :returns: ``True`` if future cost changes allowed; else ``False``.
        """
        return self.config.getbool('rattail.batch', 'vendor_catalog.allow_future',
                                   default=False)

    def should_populate(self, batch):
        # all vendor catalogs must come from data file
        return True

    def setup(self, batch, progress=None):
        self.vendor = batch.vendor
        self.products = {'upc': {}, 'vendor_code': {}}
        session = self.app.get_session(batch)
        products = session.query(model.Product)\
                          .options(orm.joinedload(model.Product.brand))\
                          .options(orm.joinedload(model.Product.costs))

        def cache(product, i):
            if product.upc:
                self.products['upc'][product.upc] = product
            cost = product.cost_for_vendor(self.vendor)
            product.vendor_cost = cost
            if cost and cost.code:
                self.products['vendor_code'][cost.code] = product

        assert self.progress_loop(cache, products.all(), progress,
                                  message="Caching products by UPC and vendor code")

    setup_populate = setup
    setup_refresh = setup

    def populate(self, batch, progress=None):
        """
        Default logic just invokes :meth:`populate_from_file()`.
        """
        return self.populate_from_file(batch, progress=progress)

    def populate_from_file(self, batch, progress=None):
        """
        Populate the given batch using data from its input file.  A
        catalog parser will be instantiated and asked to read row data
        from the file.  Each row is then added to the batch.

        The batch must have valid
        :attr:`~rattail.db.model.batch.vendorcatalog.VendorCatalogBatch.filename`
        and
        :attr:`~rattail.db.model.batch.vendorcatalog.VendorCatalogBatch.parser_key`
        attributes.  The path to the input file will be determined by
        invoking the
        :meth:`~rattail.db.model.batch.core.BatchMixin.filepath()`
        method on the batch.

        :param batch: The batch to be populated.

        :param progress: Optional progress factory.
        """
        if not batch.filename:
            raise ValueError("batch does not have a filename: {}".format(batch))
        if not batch.parser_key:
            raise ValueError("batch does not have a parser_key: {}".format(batch))

        session = self.app.get_session(batch)
        path = batch.filepath(self.config)
        vendor_handler = self.app.get_vendor_handler()
        parser = vendor_handler.get_catalog_parser(batch.parser_key,
                                                   require=True)
        parser.session = session
        parser.vendor = batch.vendor
        if not batch.effective:
            batch.effective = parser.parse_effective_date(path)

        def append(row, i):
            self.add_row(batch, row)
            if i % 500 == 0: # pragma: no cover
                session.flush()

        data = list(parser.parse_rows(path, progress=progress))
        self.progress_loop(append, data, progress,
                           message="Adding initial rows to batch")

    def identify_product(self, row):
        """
        Try to locate the product represented by the given row.
        Lookups are done using either the ``upc`` or ``vendor_code``
        attributes of the row.

        Under normal circumstances the batch handler will have
        pre-cached all existing products, for quicker lookup.  For
        instance this is the case for the full populate and refresh
        actions.  But this logic is able to do its own slower lookups
        if there is no cache available.

        :param row: A
           :class:`~rattail.db.model.batch.vendorcatalog.VendorCatalogBatchRow`
           instance.

        :returns: A :class:`~rattail.db.model.products.Product`
           instance, or ``None`` if no match could be found.
        """
        products_handler = self.app.get_products_handler()
        session = self.app.get_session(row)
        product = None

        if row.upc:
            if hasattr(self, 'products'):
                product = self.products['upc'].get(row.upc)
            else:
                product = products_handler.locate_product_for_gpc(
                    session, row.upc)

        if not product and row.vendor_code:
            if hasattr(self, 'products'):
                product = self.products['vendor_code'].get(row.vendor_code)
            else:
                product = products_handler.locate_product_for_entry(
                    session, row.vendor_code, product_key='ignore',
                    lookup_vendor_code=True, vendor=row.batch.vendor)

        return product

    def refresh_row(self, row):
        """
        Refresh data attributes and status for the given row.

        For a vendor catalog, the typical thing is done for basic
        product attributes.

        If case cost is known but unit cost is not, the latter will be
        calculated if possible.

        "Old" (i.e. "current" prior to batch execution) values will
        all be re-fetched from the main database(s), and "diff" values
        will be re-calculated.

        :param row: A
           :class:`~rattail.db.model.batch.vendorcatalog.VendorCatalogBatchRow`
           instance.
        """
        batch = row.batch

        # clear this first in case it's set
        row.status_text = None

        if not row.product:
            row.product = self.identify_product(row)
            if not row.product:
                row.status_code = row.STATUS_PRODUCT_NOT_FOUND
                return

        product = row.product

        row.upc = row.product.upc
        row.item_id = row.product.item_id
        row.brand_name = row.product.brand.name if row.product.brand else None
        row.description = row.product.description
        row.size = row.product.size

        # maybe calculate unit cost from case cost
        if row.unit_cost is None and row.case_cost is not None:
            if row.case_size is not None:
                # nb. sometimes both case size and cost are integers,
                # and simple division yields a float!  this approach
                # should hopefully work regardless, to get a decimal.
                row.unit_cost = decimal.Decimal('{:0.4f}'.format(
                    row.case_cost / row.case_size))

        if hasattr(product, 'vendor_cost'):
            old_cost = product.vendor_cost
        else:
            old_cost = product.cost_for_vendor(batch.vendor)
        if not old_cost:
            row.status_code = row.STATUS_NEW_COST
            return

        row.cost = old_cost
        row.old_vendor_code = old_cost.code
        row.old_case_size = old_cost.case_size
        row.old_case_cost = old_cost.case_cost
        row.old_unit_cost = old_cost.unit_cost

        # only consider vendor match if product does in fact have a vendor
        # TODO: at least i assume that's a reasonable idea?
        if row.product.costs:
            row.is_preferred_vendor = row.product.costs[0].vendor is row.batch.vendor

        self.refresh_cost_diffs(row)
        self.set_status_per_diffs(row)

    def refresh_cost_diffs(self, row):

        # old_case_cost
        if row.case_cost is not None and row.old_case_cost is not None:
            row.case_cost_diff = row.case_cost - row.old_case_cost

        # old_unit_cost
        if row.unit_cost is not None and row.old_unit_cost is not None:
            row.unit_cost_diff = row.unit_cost - row.old_unit_cost
            if row.old_unit_cost:
                row.unit_cost_diff_percent = row.unit_cost_diff / row.old_unit_cost
            else:
                row.unit_cost_diff_percent = 1

    def set_status_per_diffs(self, row):

        if row.vendor_code != row.old_vendor_code:
            row.status_code = row.STATUS_CHANGE_VENDOR_ITEM_CODE
            row.status_text = "new vendor item code {} differs from old code {}".format(
                repr(row.vendor_code), repr(row.old_vendor_code))
            return

        if row.case_size != row.old_case_size:
            row.status_code = row.STATUS_CHANGE_CASE_SIZE
            row.status_text = "new case size {} differs from old case size {}".format(
                repr(row.case_size), repr(row.old_case_size))
            return

        if row.case_cost != row.old_case_cost:
            diff_meets_threshold = True
            if self.case_cost_diff_threshold and self.case_cost_diff_threshold > abs(
                    row.case_cost - row.old_case_cost):
                diff_meets_threshold = False
            if diff_meets_threshold:
                row.status_code = row.STATUS_CHANGE_COST
                row.status_text = "new case cost {} differs from old cost {}".format(
                    repr(row.case_cost), repr(row.old_case_cost))
                return

        if row.unit_cost != row.old_unit_cost:
            diff_meets_threshold = True
            if (self.unit_cost_diff_threshold
                and row.unit_cost is not None and row.old_unit_cost is not None
                and self.unit_cost_diff_threshold > abs(
                    row.unit_cost - row.old_unit_cost)):
                diff_meets_threshold = False
            if diff_meets_threshold:
                row.status_code = row.STATUS_CHANGE_COST
                row.status_text = "new unit cost {} differs from old cost {}".format(
                    repr(row.unit_cost), repr(row.old_unit_cost))
                return

        row.status_code = row.STATUS_NO_CHANGE

    # TODO: who uses this..?
    def cost_differs(self, row, cost):
        """
        Compare a batch row with a cost record to determine whether they match
        or differ.
        """
        if row.vendor_code is not None and row.vendor_code != cost.code:
            return "new vendor code {} differs from old code {}".format(
                repr(row.vendor_code), repr(cost.code))
        if row.case_cost is not None and row.case_cost != cost.case_cost:
            return "new case cost {} differs from old cost {}".format(
                row.case_cost, cost.case_cost)
        if row.unit_cost is not None and row.unit_cost != cost.unit_cost:
            return "new unit cost {} differs from old cost {}".format(
                row.unit_cost, cost.unit_cost)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data Models for Stores
"""

from __future__ import unicode_literals, absolute_import

import six
import sqlalchemy as sa
from sqlalchemy.orm import relationship
from sqlalchemy.ext.orderinglist import ordering_list

from .core import Base, uuid_column
from .contact import PhoneNumber, EmailAddress


@six.python_2_unicode_compatible
class Store(Base):
    """
    Represents a store (physical or otherwise) within the organization.
    """
    __tablename__ = 'store'
    __versioned__ = {}

    uuid = uuid_column()

    id = sa.Column(sa.String(length=10), nullable=True, doc="""
    Supposedly unique ID string for the store.
    """)

    name = sa.Column(sa.String(length=100), nullable=True, doc="""
    Common name for the store.
    """)

    database_key = sa.Column(sa.String(length=30), nullable=True, doc="""
    Supposedly unique "config key" for the store.  This key is used e.g. to
    define the database connection for the store.
    """)

    archived = sa.Column(sa.Boolean(), nullable=True, doc="""
    If set, this should indicate the record is "archived" and is no longer in
    use per se.
    """)

    def __str__(self):
        return str(self.name or '')

    def add_email_address(self, address, type='Info'):
        email = StoreEmailAddress(address=address, type=type)
        self.emails.append(email)

    def add_phone_number(self, number, type='Voice'):
        phone = StorePhoneNumber(number=number, type=type)
        self.phones.append(phone)

    # TODO: this class should use ContactMixin instead
    def first_phone(self, **kwargs):
        if self.phones:
            return self.phones[0]


class StorePhoneNumber(PhoneNumber):
    """
    Represents a phone (or fax) number associated with a store.
    """

    __mapper_args__ = {'polymorphic_identity': 'Store'}


Store.phones = relationship(
    StorePhoneNumber,
    backref='store',
    primaryjoin=StorePhoneNumber.parent_uuid == Store.uuid,
    foreign_keys=[StorePhoneNumber.parent_uuid],
    collection_class=ordering_list('preference', count_from=1),
    order_by=StorePhoneNumber.preference,
    cascade='save-update, merge, delete, delete-orphan')

Store.phone = relationship(
    StorePhoneNumber,
    primaryjoin=sa.and_(
        StorePhoneNumber.parent_uuid == Store.uuid,
        StorePhoneNumber.preference == 1),
    foreign_keys=[StorePhoneNumber.parent_uuid],
    uselist=False,
    viewonly=True)


class StoreEmailAddress(EmailAddress):
    """
    Represents an email address associated with a store.
    """

    __mapper_args__ = {'polymorphic_identity': 'Store'}


Store.emails = relationship(
    StoreEmailAddress,
    backref='store',
    primaryjoin=StoreEmailAddress.parent_uuid == Store.uuid,
    foreign_keys=[StoreEmailAddress.parent_uuid],
    collection_class=ordering_list('preference', count_from=1),
    order_by=StoreEmailAddress.preference,
    cascade='save-update, merge, delete, delete-orphan')

Store.email = relationship(
    StoreEmailAddress,
    primaryjoin=sa.and_(
        StoreEmailAddress.parent_uuid == Store.uuid,
        StoreEmailAddress.preference == 1),
    foreign_keys=[StoreEmailAddress.parent_uuid],
    uselist=False,
    viewonly=True)
